@@ -15,18 +15,26 @@ use anyhow::Result;
 use sequoia_openpgp as openpgp;
 use openpgp::{
     Fingerprint,
-    KeyID,
+    KeyHandle,
+    cert::prelude::*,
     crypto::hash::Digest,
     crypto::mpi::PublicKey,
-    packet::{UserID, Key, key::{PublicParts, PrimaryRole, SubordinateRole}},
+    packet::{
+        Key,
+        Signature,
+        UserID,
+        key::{PublicParts, PrimaryRole, SubordinateRole},
+        signature::subpacket::SubpacketTag,
+    },
     types::*,
 };
 use sequoia_ipc as ipc;
 use ipc::Keygrip;
 
 use crate::{
+    KeyIDFormat,
     babel,
-    common::{Common, Compliance},
+    common::{Common, Compliance, PublicKeyAlgorithmAndSize, get_curve},
     trust::*,
 };
 
@@ -57,24 +65,14 @@ pub enum Record<'k> {
     Fingerprint(Fingerprint),
     Keygrip(Keygrip),
     UserID {
+        amalgamation: UserIDAmalgamation<'k>,
         validity: Option<Validity>,
-        creation_date: SystemTime,
-        expiration_date: Option<SystemTime>,
-        userid: UserID,
     },
 
     Signature {
-        issuer: Option<KeyID>,
-        issuer_fp: Option<Fingerprint>,
-        issuer_uid: Option<UserID>,
-        validity: Option<SignatureValidity>,
-        pk_algo: PublicKeyAlgorithm,
-        hash_algo: HashAlgorithm,
-        creation_time: SystemTime,
-        typ: SignatureType,
-        exportable: bool,
-        trust: Option<(u8, u8)>,
-        has_notations: bool,
+        sig: &'k Signature,
+        issuer_uid: IssuerUserID,
+        validity: SignatureValidity,
     },
 
     /// rvk: Revocation key
@@ -206,10 +204,22 @@ impl Record<'_> {
                              compliance_flags,
                     )?;
                 } else {
+                    let algo = if let Some(c) = curve {
+                        PublicKeyAlgorithmAndSize::Ecc(c)
+                    } else {
+                        PublicKeyAlgorithmAndSize::VariableLength(
+                            key.pk_algo(), key_length)
+                    };
+
                     writeln!(w,
-                             "{}   {} {} [{}]{}",
+                             "{}   {}{} {} [{}]{}",
                              record_type,
-                             babel::Fish((key.pk_algo(), key_length, &curve)),
+                             babel::Fish(algo),
+                             match config.keyid_format {
+                                 KeyIDFormat::None => format!(""),
+                                 KeyIDFormat::Long => format!("/{}", key.keyid()),
+                                 KeyIDFormat::HexLong => format!("/0x{}", key.keyid()),
+                             },
                              creation_date.format("%Y-%m-%d"),
                              babel::Fish(primary_key_flags).to_string().to_uppercase(),
                              bracket(config, revocation_date, expiration_date),
@@ -271,10 +281,22 @@ impl Record<'_> {
                              compliance_flags,
                     )?;
                 } else {
+                    let algo = if let Some(c) = curve {
+                        PublicKeyAlgorithmAndSize::Ecc(c)
+                    } else {
+                        PublicKeyAlgorithmAndSize::VariableLength(
+                            key.pk_algo(), key_length)
+                    };
+
                     writeln!(w,
-                             "{}   {} {} [{}]{}",
+                             "{}   {}{} {} [{}]{}",
                              record_type,
-                             babel::Fish((key.pk_algo(), key_length, &curve)),
+                             babel::Fish(algo),
+                             match config.keyid_format {
+                                 KeyIDFormat::None => format!(""),
+                                 KeyIDFormat::Long => format!("/{}", key.keyid()),
+                                 KeyIDFormat::HexLong => format!("/0x{}", key.keyid()),
+                             },
                              creation_date.format("%Y-%m-%d"),
                              babel::Fish(key_flags).to_string().to_uppercase(),
                              bracket(config, revocation_date, expiration_date),
@@ -303,13 +325,64 @@ impl Record<'_> {
             },
 
             UserID {
+                amalgamation,
                 validity,
-                creation_date,
-                expiration_date,
-                userid,
             } => {
+                let userid = amalgamation.userid();
+                let binding = amalgamation
+                    .binding_signature(config.policy(), config.now()).ok()
+                    .or_else(|| amalgamation.self_signatures().next());
+
+                let creation_date = binding.as_ref()
+                    .and_then(|b| b.signature_creation_time())
+                    .unwrap_or(std::time::UNIX_EPOCH);
+
+                let expiration_date = binding.as_ref()
+                    .and_then(|b| b.signature_expiration_time());
+
+                // GnuPG doesn't emit the creation date if the user ID
+                // binding signature is expired or revoked.
                 let creation_date =
-                    chrono::DateTime::<chrono::Utc>::from(*creation_date);
+                    (
+                        match amalgamation.revocation_status(
+                            config.policy(), config.now())
+                        {
+                            // Never emit for revoked user IDs.
+                            RevocationStatus::Revoked(_) => false,
+
+                            RevocationStatus::CouldBe(revs) => {
+                                use openpgp::cert::amalgamation::ValidAmalgamation;
+                                let revokers = amalgamation.clone().
+                                    with_policy(config.policy(), config.now())
+                                    .map(|vca| vca.revocation_keys()
+                                         .map(|r| {
+                                             let (pk, fp) = r.revoker();
+                                             (pk, KeyHandle::from(fp))
+                                         })
+                                         .collect::<Vec<_>>())
+                                    .unwrap_or_default();
+
+                                // Never emit for third-party revoked
+                                // user IDs.
+                                //
+                                // XXX: We don't validate the
+                                // signature.
+                                ! revs.iter().any(|rev| {
+                                    let issuers = rev.get_issuers();
+                                    revokers.iter().any(|(pk, kh)| {
+                                        *pk == rev.pk_algo()
+                                            && issuers.contains(kh)
+                                    })
+                                })
+                            },
+
+                            // Always emit for non-revoked user IDs.
+                            RevocationStatus::NotAsFarAsWeKnow => true,
+                        }
+                        && expiration_date.is_none()
+                    ).then_some(
+                        chrono::DateTime::<chrono::Utc>::from(creation_date)
+                    );
 
                 let expiration_date = expiration_date.map(|t| {
                     chrono::DateTime::<chrono::Utc>::from(t)
@@ -322,7 +395,8 @@ impl Record<'_> {
                 if mr {
                     write!(w, "uid:{}::::{}:{}:{}::",
                            validity.unwrap_or(ValidityLevel::Unknown.into()),
-                           creation_date.format("%s"),
+                           creation_date.map(|t| t.format("%s").to_string())
+                           .unwrap_or_default(),
                            expiration_date.map(|t| t.format("%s").to_string())
                            .unwrap_or_else(|| "".into()),
                            openpgp::fmt::hex::encode(&uidhash),
@@ -331,9 +405,16 @@ impl Record<'_> {
                     writeln!(w, "::::::::::0:")?;
                 } else {
                     if let Some(validity) = validity {
-                        writeln!(w, "uid           {} {}",
+                        writeln!(w, "{:width$}{} {}",
+                                 "uid",
                                  BoxedValidity(*validity),
-                                 String::from_utf8_lossy(userid.value()))?;
+                                 String::from_utf8_lossy(userid.value()),
+                                 width = match config.keyid_format {
+                                     KeyIDFormat::None => 14,
+                                     KeyIDFormat::Long => 20,
+                                     KeyIDFormat::HexLong => 22,
+                                 },
+                        )?;
                     } else {
                         writeln!(w, "uid                      {}",
                                  String::from_utf8_lossy(userid.value()))?;
@@ -342,19 +423,25 @@ impl Record<'_> {
             },
 
             Signature {
-                issuer,
-                issuer_fp,
+                sig,
                 issuer_uid,
                 validity,
-                pk_algo,
-                hash_algo,
-                creation_time,
-                typ,
-                exportable,
-                trust,
-                has_notations,
             } => {
                 use SignatureType::*;
+                let issuer =  sig.issuers().cloned().next().or_else(
+                    || sig.issuer_fingerprints().cloned().next().map(Into::into));
+                let issuer_fp = sig.issuer_fingerprints().cloned().next();
+
+                let pk_algo = sig.pk_algo();
+                let hash_algo = sig.hash_algo();
+                let creation_time = sig.signature_creation_time()
+                    .expect("valid signatures have a creation time");
+                let typ = sig.typ();
+                let exportable = sig.exportable().is_ok();
+                let trust = sig.trust_signature();
+                let has_notations = sig.subpackets(SubpacketTag::NotationData)
+                    .next().is_some();
+
                 let class = match typ {
                     CertificationRevocation
                         | KeyRevocation
@@ -363,49 +450,98 @@ impl Record<'_> {
                 };
 
                 let creation_time =
-                    chrono::DateTime::<chrono::Utc>::from(*creation_time);
+                    chrono::DateTime::<chrono::Utc>::from(creation_time);
 
                 if mr {
-                    writeln!(w, "{}:{}::{}:{}:{}::{}::{}:{:02x}{}::{}:::{}:",
+                    // For some reason, GnuPG suppresses this in colon-mode.
+                    let issuer_uid =
+                        if issuer_uid == &IssuerUserID::NotFound && config.check_sigs {
+                            &IssuerUserID::Empty
+                        } else {
+                            issuer_uid
+                        };
+
+                    writeln!(w, "{}:{:#}::{}:{}:{}::{}::{}:{:02x}{}::{}:::{}:",
                              class,
-                             validity.as_ref().map(|i| i.to_string())
-                             .unwrap_or_default(),
-                             u8::from(*pk_algo),
+                             validity,
+                             u8::from(pk_algo),
                              issuer.as_ref().map(|i| i.to_string())
                              .unwrap_or_default(),
                              creation_time.format("%s"),
                              trust.map(|(depth, amount)|
                                        format!("{} {}", depth, amount))
                              .unwrap_or_else(|| "".into()),
-                             issuer_uid.as_ref()
-                             .map(|u| String::from_utf8_lossy(u.value()).to_string())
-                             .unwrap_or_else(|| "[User ID not found]".to_string()),
-                             u8::from(*typ),
-                             if *exportable { 'x' } else { 'l' },
+                             issuer_uid,
+                             u8::from(typ),
+                             if exportable { 'x' } else { 'l' },
                              issuer_fp.as_ref().map(|i| i.to_string())
                              .unwrap_or_default(),
-                             u8::from(*hash_algo))?;
+                             u8::from(hash_algo))?;
                 } else {
                     use SignatureType::*;
-                    writeln!(w, "{} {}    {} {} {} {}  {}",
+                    writeln!(w, "{}{}{}   {}{} {} {} {}  {}",
                              class,
+                             validity,
                              match typ {
                                  PersonaCertification => '1',
                                  CasualCertification => '2',
                                  PositiveCertification => '3',
                                  _ => ' ',
                              },
+                             sig.policy_uri().map(|_| 'P').unwrap_or(' '),
                              has_notations.then_some('N').unwrap_or(' '),
                              trust.map(|(depth, _amount)| depth.to_string())
                              .unwrap_or_else(|| " ".into()),
                              issuer.as_ref().map(|i| i.to_string())
                              .unwrap_or_default(),
                              creation_time.format("%Y-%m-%d"),
-                             issuer_uid.as_ref()
-                             .map(|u|
-                                  String::from_utf8_lossy(u.value()).to_string())
-                             .unwrap_or_else(
-                                 || "[User ID not found]".to_string()))?;
+                             issuer_uid)?;
+                    if let Some(p) = sig.policy_uri()
+                        .filter(|_| config.list_options.policy_urls)
+                    {
+                        writeln!(w, "   Signature policy: {}",
+                                 String::from_utf8_lossy(p))?;
+                    }
+
+                    let lc = &config.list_options;
+                    if lc.ietf_notations || lc.user_notations {
+                        let notations: Vec<_> =
+                            sig.notation_data()
+                            .filter(|n| {
+                                let user = n.name().contains('@');
+                                (! user && lc.ietf_notations)
+                                    || (user && lc.user_notations)
+                            })
+                            .collect();
+
+                        for n in notations.iter().rev() {
+                            writeln!(w, "   Signature notation: {}={}",
+                                     n.name(),
+                                     if n.flags().human_readable() {
+                                         String::from_utf8_lossy(n.value())
+                                             .to_string()
+                                     } else {
+                                         let preview: String =
+                                             n.value().iter().take(19)
+                                             .map(|b|
+                                                  char::from_u32((*b).into())
+                                                  .filter(|c| *c == ' '
+                                                          || c.is_ascii_graphic())
+                                                  .unwrap_or('?'))
+                                             .collect();
+
+                                         format!("[ not human readable \
+                                                  ({} bytes: {}{}) ]",
+                                                 n.value().len(),
+                                                 preview,
+                                                 if n.value().len() > 20 {
+                                                     "..."
+                                                 } else {
+                                                     ""
+                                                 })
+                                     })?;
+                        }
+                    }
                 }
             },
 
@@ -480,7 +616,9 @@ impl fmt::Display for TokenSN {
     }
 }
 
+#[derive(Debug, PartialEq, Eq)]
 pub enum SignatureValidity {
+    NotChecked,
     Good,
     Bad,
     MissingKey,
@@ -491,14 +629,57 @@ impl fmt::Display for SignatureValidity {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         use SignatureValidity::*;
         match self {
+            NotChecked => if f.alternate() {
+                // Suppress output for machine-readable interface.
+                Ok(())
+            } else {
+                f.write_str(" ")
+            },
             Good => f.write_str("!"),
             Bad => f.write_str("-"),
-            MissingKey => f.write_str("?"),
+            MissingKey =>  if f.alternate() {
+                f.write_str("?")
+            } else {
+                // Suppress output for human-readable interface.
+                f.write_str(" ")
+            },
             OtherError => f.write_str("%"),
         }
     }
 }
 
+#[derive(Debug, PartialEq, Eq)]
+pub enum IssuerUserID {
+    NotFound,
+    TimeConflict,
+    Empty,
+    Some(String),
+}
+
+impl From<&UserID> for IssuerUserID {
+    fn from(u: &UserID) -> IssuerUserID {
+        IssuerUserID::Some(String::from_utf8_lossy(u.value()).into())
+    }
+}
+
+impl From<String> for IssuerUserID {
+    fn from(u: String) -> IssuerUserID {
+        IssuerUserID::Some(u)
+    }
+}
+
+impl fmt::Display for IssuerUserID {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        use IssuerUserID::*;
+        match self {
+            NotFound => f.write_str("[User ID not found]"),
+            TimeConflict => f.write_str("[Time conflict] "),
+            Empty => Ok(()),
+            Some(u) => f.write_str(&u),
+        }
+    }
+}
+
 /// Escapes the given string.
 pub fn escape(s: impl AsRef<str>) -> String {
     e_str(s.as_ref())
@@ -563,16 +744,6 @@ impl fmt::Display for BoxedValidity {
     }
 }
 
-/// Returns the elliptic curve of the given key, if any.
-pub fn get_curve(mpis: &PublicKey) -> Option<Curve> {
-    match mpis {
-        PublicKey::EdDSA { curve, .. }
-        | PublicKey::ECDSA { curve, .. }
-        | PublicKey::ECDH { curve, .. } => Some(curve.clone()),
-        _ => None,
-    }
-}
-
 /// Returns the size of the key that we should report.
 pub fn get_bits(mpis: &PublicKey) -> usize {
     match mpis {
