@@ -13,8 +13,10 @@ use sequoia_openpgp as openpgp;
 use openpgp::{
     Fingerprint,
     KeyHandle,
-    packet::UserID,
+    crypto::mpi::PublicKey,
+    packet::{Key, UserID, key},
     policy::Policy,
+    types::{Curve, PublicKeyAlgorithm},
 };
 
 use sequoia_cert_store as cert_store;
@@ -59,7 +61,13 @@ pub trait Common<'store> {
     ///
     /// In contrast to Self::warn, this makes the program report a
     /// failure when exiting.
-    fn error(&self, msg: fmt::Arguments);
+    fn error(&self, msg: fmt::Arguments) {
+        self.warn(msg);
+        self.fail();
+    }
+
+    /// Makes the program report a failure when exiting.
+    fn fail(&self);
 
     /// Sets an explicit status code, and prevents the error message
     /// to be shown at the end of the main function.
@@ -69,7 +77,7 @@ pub trait Common<'store> {
 
     /// Prints the usage and the given message and returns an error.
     fn wrong_args(&self, message: &str) -> Result<()> {
-        eprintln!("usage: {} [options] {}", self.argv0(), message);
+        safe_eprintln!("usage: {} [options] {}", self.argv0(), message);
         self.override_status_code(2);
         Err(anyhow::anyhow!("Wrong arguments: {}", message))
     }
@@ -129,15 +137,19 @@ pub trait Common<'store> {
     fn with_fingerprint(&self) -> bool;
 }
 
-#[derive(Clone, Copy, Debug, PartialEq, Eq)]
+#[derive(Clone, Copy, Debug, Default, PartialEq, Eq)]
 pub enum TrustModel {
+    #[default]
     PGP,
+    GnuPG,
     Classic,
     Always,
     Direct,
     Tofu,
     TofuPGP,
     Auto,
+    Sequoia,
+    SequoiaGnuPG,
     Unknown(u8),
 }
 
@@ -153,29 +165,29 @@ impl From<u8> for TrustModel {
             5 => Auto,
             6 => Tofu,
             7 => TofuPGP,
+            253 => GnuPG,
+            254 => SequoiaGnuPG,
+            255 => Sequoia,
             n => Unknown(n),
         }
     }
 }
 
-impl Default for TrustModel {
-    fn default() -> Self {
-        TrustModel::Auto
-    }
-}
-
 impl std::str::FromStr for TrustModel {
     type Err = anyhow::Error;
 
     fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
         match s.to_lowercase().as_str() {
             "pgp" => Ok(TrustModel::PGP),
+            "gnupg" => Ok(TrustModel::GnuPG),
             "classic" => Ok(TrustModel::Classic),
             "always" => Ok(TrustModel::Always),
             "direct" => Ok(TrustModel::Direct),
             "tofu" => Ok(TrustModel::Tofu),
             "tofu+pgp" => Ok(TrustModel::TofuPGP),
             "auto" => Ok(TrustModel::Auto),
+            "sequoia" => Ok(TrustModel::Sequoia),
+            "sequoia+gnupg" => Ok(TrustModel::SequoiaGnuPG),
             _ => Err(anyhow::anyhow!("Unknown trust model {:?}", s)),
         }
     }
@@ -186,29 +198,57 @@ impl fmt::Display for TrustModel {
         use TrustModel::*;
         match self {
             PGP => f.write_str("pgp"),
+            GnuPG => f.write_str("gnupg"),
             Classic => f.write_str("classic"),
             Always => f.write_str("always"),
             Direct => f.write_str("direct"),
             Tofu => f.write_str("tofu"),
             TofuPGP => f.write_str("tofu+pgp"),
             Auto => f.write_str("auto"),
+            Sequoia => f.write_str("sequoia"),
+            SequoiaGnuPG => f.write_str("sequoia+gnupg"),
             Unknown(n) => write!(f, "unknown({})", n),
         }
     }
 }
 
-pub trait Model {
-    fn with_policy<'a, 'store>(&self, config: &'a crate::Config<'store>,
-                               at: Option<SystemTime>)
+pub trait Model: std::fmt::Debug {
+    /// Creates a view under the given configuration, policy, and
+    /// time.
+    fn with_policy<'a, 'store>(
+        &self,
+        config: &'a crate::Config<'store>,
+        at: Option<SystemTime>)
         -> Result<Box<dyn ModelViewAt<'a, 'store> + 'a>>
-        where 'store: 'a;
+    where
+        'store: 'a,
+    {
+        self.with_policy_and_precompute(config, at, false)
+    }
+
+    /// Creates a view under the given configuration, policy, and
+    /// time, and possibly pre-compute the network.
+    ///
+    /// Note: when considering all or most of the certificates,
+    /// pre-computing will improve performance.
+    fn with_policy_and_precompute<'a, 'store>(
+        &self,
+        config: &'a crate::Config<'store>,
+        at: Option<SystemTime>,
+        precompute: bool)
+        -> Result<Box<dyn ModelViewAt<'a, 'store> + 'a>>
+    where
+        'store: 'a;
 }
 
 pub fn null_model() -> Box<dyn Model> {
+    #[derive(Debug)]
     struct Null(());
     impl Model for Null {
-        fn with_policy<'a, 'store>(&self, _: &'a crate::Config,
-                                   _: Option<SystemTime>)
+        fn with_policy_and_precompute<'a, 'store>(
+            &self, _: &'a crate::Config,
+            _: Option<SystemTime>,
+            _: bool)
             -> Result<Box<dyn ModelViewAt<'a, 'store> + 'a>>
             where 'store: 'a
         {
@@ -353,20 +393,29 @@ impl fmt::Display for Query {
     }
 }
 
-impl From<&str> for Query {
-    fn from(s: &str) -> Query {
+impl std::str::FromStr for Query {
+    type Err = anyhow::Error;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
         if s.ends_with("!") {
             if let Ok(h) = s[..s.len()-1].parse() {
-                return Query::ExactKey(h);
+                return Ok(Query::ExactKey(h));
             }
         }
 
         if let Ok(h) = s.parse() {
-            Query::Key(h)
+            if let KeyHandle::Fingerprint(Fingerprint::Invalid(b)) = &h {
+                if b.len() == 4 {
+                    return Err(anyhow::anyhow!(
+                        "Short key IDs are insecure and not supported: {}", s));
+                }
+            }
+
+            Ok(Query::Key(h))
         } else if s.starts_with("<") && s.ends_with(">") {
-            Query::Email(s[1..s.len()-1].to_lowercase())
+            Ok(Query::Email(s[1..s.len()-1].to_lowercase()))
         } else {
-            Query::UserIDFragment(s.to_lowercase())
+            Ok(Query::UserIDFragment(s.to_lowercase()))
         }
     }
 }
@@ -376,6 +425,17 @@ impl Query {
     ///
     /// Note: the match must be authenticated!
     pub fn matches(&self, cert: &Arc<LazyCert>) -> bool {
+        // We do the test twice.  First, potentially on the RawCert,
+        // where the binding signatures haven't been checked.  Only if
+        // that matches do we canonicalize the cert, and re-do the
+        // check to make sure.
+        self.matches_internal(cert)
+            && cert.to_cert()
+            .map(|_| self.matches_internal(cert))
+            .unwrap_or(false)
+    }
+
+    fn matches_internal(&self, cert: &Arc<LazyCert>) -> bool {
         match self {
             Query::Key(h) | Query::ExactKey(h) =>
                 cert.keys().any(|k| k.key_handle().aliases(h)),
@@ -407,6 +467,29 @@ impl Query {
         }
     }
 
+    /// Returns the cert's User ID matching the query.
+    ///
+    /// This falls back to a best-effort heuristic to compute the
+    /// primary User ID if the query matches a key.
+    pub fn best_effort_uid(&self,
+                           policy: &dyn Policy,
+                           cert: &openpgp::Cert)
+                           -> String
+    {
+        match self {
+            Query::Key(_) | Query::ExactKey(_) => (),
+            Query::Email(_) | Query::UserIDFragment(_) =>
+                for uidb in cert.userids() {
+                    if self.matches_userid(uidb.userid()) {
+                        return String::from_utf8_lossy(
+                            uidb.userid().value()).into();
+                    }
+                },
+        }
+
+        crate::utils::best_effort_primary_uid(policy, cert)
+    }
+
     /// Returns whether this query uses a fingerprint or key ID.
     pub fn by_key_handle(&self) -> bool {
         match self {
@@ -589,3 +672,41 @@ impl Compliance {
         }
     }
 }
+
+/// Returns the elliptic curve of the given key, if any.
+pub fn get_curve(mpis: &PublicKey) -> Option<Curve> {
+    match mpis {
+        PublicKey::EdDSA { curve, .. }
+        | PublicKey::ECDSA { curve, .. }
+        | PublicKey::ECDH { curve, .. } => Some(curve.clone()),
+        _ => None,
+    }
+}
+
+/// Public key algorithm and optional key length.
+#[derive(Debug)]
+pub enum PublicKeyAlgorithmAndSize {
+    /// Public key algorithm with variable-sized keys.
+    VariableLength(PublicKeyAlgorithm, usize),
+
+    /// Public key algorithm with fixed-size keys.
+    FixedLength(PublicKeyAlgorithm),
+
+    /// ECC algorithm with fixed-size keys.
+    Ecc(Curve),
+}
+
+impl<P, R> From<&Key<P, R>> for PublicKeyAlgorithmAndSize
+where
+    P: key::KeyParts,
+    R: key::KeyRole,
+{
+    fn from(key: &Key<P, R>) -> PublicKeyAlgorithmAndSize {
+        if let Some(c) = get_curve(key.mpis()) {
+            PublicKeyAlgorithmAndSize::Ecc(c)
+        } else {
+            PublicKeyAlgorithmAndSize::VariableLength(
+                key.pk_algo(), key.mpis().bits().unwrap_or(0))
+        }
+    }
+}
