@@ -19,11 +19,17 @@ use openpgp::{
     serialize::{Serialize, stream::*},
     types::*,
 };
-use sequoia_ipc as ipc;
 
 use crate::{
     babel,
-    common::{Common, Query, TrustModel, Validity, ValidityLevel},
+    common::{
+        Common,
+        PublicKeyAlgorithmAndSize,
+        Query,
+        TrustModel,
+        Validity,
+        ValidityLevel,
+    },
     compliance::Compliance,
     status::{self, Status, InvalidKeyReason},
     utils,
@@ -126,7 +132,7 @@ fn do_encrypt(config: &crate::Config, args: &[String],
 
     for recipient in &config.remote_user {
         // XXX: honor constraints
-        let query = crate::trust::Query::from(recipient.name.as_str());
+        let query: crate::trust::Query = recipient.name.parse()?;
 
         // XXX: One remote user may expand to multiple recipients.  In
         // the case of groups, this is a feature.  In the case of
@@ -243,7 +249,7 @@ fn do_encrypt(config: &crate::Config, args: &[String],
     }
 
     let recipients: Vec<Recipient>
-        = keys.iter().map(Recipient::from).collect();
+        = keys.iter().rev().map(Recipient::from).collect();
 
     let mut sink = if let Some(name) = outfile {
         utils::create(config, name)?
@@ -265,7 +271,23 @@ fn do_encrypt(config: &crate::Config, args: &[String],
     // The reason is that we want to produce the SKESK ourselves so
     // that we can cache the password in the agent.  To that end, fix
     // cipher and session key here.
-    let cipher = config.def_cipher;
+
+    let cipher = if let Some(def_cipher) = config.def_cipher {
+        // This overrides the recipients preferences, but we may warn
+        // about that, unless we deem ourselves an expert.
+        if ! config.expert && ! cipher_preferences.contains(&def_cipher) {
+            config.warn(format_args!(
+                "WARNING: forcing symmetric cipher {} ({}) \
+                 violates recipient preferences",
+                babel::Fish(def_cipher), u8::from(def_cipher)));
+        }
+
+        def_cipher
+    } else {
+        // Select best cipher from the recipient's preferences.
+        cipher_preferences.get(0).cloned().unwrap_or_default()
+    };
+
     let sk = SessionKey::new(cipher.key_size()?);
     de_vs_compliant &=
         config.de_vs_producer.symmetric_algorithm(cipher).is_ok();
@@ -273,7 +295,7 @@ fn do_encrypt(config: &crate::Config, args: &[String],
     // Now do our trick, maybe.
     if symmetric {
         let s2k = S2K::default();
-        let cacheid = crate::agent::cacheid_of(&s2k);
+        let cacheid = crate::gpg_agent::cacheid_of(&s2k);
 
         let p = rt.block_on(ask_password(config, cacheid))?;
         // XXX: We emit the SKESK first.  Naive consumers may
@@ -288,20 +310,15 @@ fn do_encrypt(config: &crate::Config, args: &[String],
         de_vs_compliant &= recipients.is_empty();
     }
 
-    if ! cipher_preferences.contains(&cipher) {
-        config.warn(format_args!(
-            "WARNING: forcing symmetric cipher {} ({}) \
-             violates recipient preferences",
-            babel::Fish(cipher), u8::from(cipher)));
-    }
-
     let encryptor = Encryptor2::with_session_key(message, cipher, sk)?
         .add_recipients(recipients);
 
     let mut message = encryptor.build()?;
 
-    if config.compress_algo != CompressionAlgorithm::Uncompressed {
-        message = Compressor::new(message).algo(config.compress_algo).build()?;
+    if let Some(algo) = config.compress_algo
+        .filter(|&a| a != CompressionAlgorithm::Uncompressed)
+    {
+        message = Compressor::new(message).algo(algo).build()?;
     }
 
     if sign {
@@ -310,7 +327,7 @@ fn do_encrypt(config: &crate::Config, args: &[String],
             rt.block_on(crate::sign::get_signers(config))?;
 
         let timestamp = config.now().try_into()?;
-        let hash_algo = config.def_digest;
+        let hash_algo = config.def_digest.unwrap_or_default();
         let mut signer =
             Signer::new(message, signers.pop().expect("at least one"))
             .creation_time(timestamp)
@@ -361,26 +378,26 @@ fn do_encrypt(config: &crate::Config, args: &[String],
 
 async fn ask_password(config: &crate::Config<'_>, cacheid: Option<String>)
                       -> Result<Password> {
+    use sequoia_gpg_agent::PinentryMode;
     let mut agent = config.connect_agent().await?;
-    Ok(crate::agent::get_passphrase(
+
+    if matches!(config.pinentry_mode, PinentryMode::Loopback)
+        && config.static_passphrase.borrow().is_none()
+    {
+        // GnuPG emits this twice, for good measure.  The second time
+        // we emit it from Config::get_passphrase.
+        config.status().emit(Status::InquireMaxLen(100))?;
+    }
+
+    Ok(config.get_passphrase(
         &mut agent,
         &cacheid, &None, None, None, false, 0, false, false,
-        |_agent, response| if let ipc::assuan::Response::Inquire {
-            keyword, parameters } = response
-        {
-            match keyword.as_str() {
-                "PINENTRY_LAUNCHED" => {
-                    let p = parameters.unwrap_or_default();
-                    let info = String::from_utf8_lossy(&p);
-                    let _ = config.status().emit(
-                        Status::PinentryLaunched(info.into()));
-                    None
-                },
-                _ => None,
-            }
-        } else {
-            None
-        }
+        |p| {
+            let info = String::from_utf8_lossy(&p);
+            config.status().emit(
+                Status::PinentryLaunched(info.into()))?;
+            Ok(())
+        },
     ).await?)
 }
 
@@ -451,12 +468,10 @@ fn do_we_trust(config: &crate::Config,
         let cert_fp = cert.fingerprint();
         let primary = fp == cert_fp;
 
-        eprintln!();
-        eprintln!("{}  {}/{} {} {}",
+        safe_eprintln!();
+        safe_eprintln!("{}  {}/{} {} {}",
                   if primary { "pub" } else { "sub" },
-                  babel::Fish((key.pk_algo(),
-                               key.mpis().bits().unwrap_or_default(),
-                               &crate::colons::get_curve(key.mpis()))),
+                  babel::Fish(PublicKeyAlgorithmAndSize::from(key)),
                   KeyID::from(&fp),
                   {
                       let creation_date =
@@ -464,26 +479,26 @@ fn do_we_trust(config: &crate::Config,
                               key.creation_time());
                       creation_date.format("%Y-%m-%d")
                   },
-                  utils::best_effort_uid_for_query(config.policy(), cert, query));
+                  query.best_effort_uid(config.policy(), cert));
 
-        eprintln!(" Primary key fingerprint: {}", cert_fp.to_spaced_hex());
+        safe_eprintln!(" Primary key fingerprint: {}", cert_fp.to_spaced_hex());
         if ! primary {
-            eprintln!("      Subkey fingerprint: {}", fp.to_spaced_hex());
+            safe_eprintln!("      Subkey fingerprint: {}", fp.to_spaced_hex());
         }
-        eprintln!();
+        safe_eprintln!();
 
         if validity.level == ValidityLevel::Never {
-            eprintln!(
+            safe_eprintln!(
                 "This key is bad!  It has been marked as untrusted!  If you\n\
                  *really* know what you are doing, you may answer the next\n\
                  question with yes.");
         } else {
-            eprintln!(
+            safe_eprintln!(
                 "It is NOT certain that the key belongs to the person named\n\
                  in the user ID.  If you *really* know what you are doing,\n\
                  you may answer the next question with yes.");
         }
-        eprintln!();
+        safe_eprintln!();
 
         config.status().emit(
             Status::UserIdHint {
