@@ -163,7 +163,7 @@ async fn real_cmd_quick_add_key(config: &mut crate::Config<'_>, args: &[String])
                   pk_length,
                   curve,
                   config.now(),
-                  crate::utils::parse_expiration(config, &expire)?,
+                  crate::utils::parse_expiration(config.now(), &expire)?,
                   usage,
                   None,
                   &config.def_preferences.clone(),
@@ -191,11 +191,9 @@ async fn real_cmd_quick_add_key(config: &mut crate::Config<'_>, args: &[String])
 
     // Store the secrets in the agent.
     let mut agent = config.connect_agent().await?;
-    // See if we import a new key or subkey.
-    crate::agent::import(&mut agent,
-                         config.policy(),
-                         &cert, &subkey,
-                         config.batch).await?;
+    agent.import(config.policy(),
+                 &cert, &subkey,
+                 config.batch, false).await?;
 
     Ok(())
 }
@@ -554,7 +552,7 @@ async fn proc_parameter_file<'a>(config: &mut crate::Config<'_>,
 	    "name-email" => Parameter::NameEmail(value.into()),
 	    "name-comment" => Parameter::NameComment(value.into()),
 	    "expire-date" => Parameter::ExpireDate(
-                crate::utils::parse_expiration(config, value)
+                crate::utils::parse_expiration(config.now(), value)
                     .map_err(|_| error_out("invalid expire date"))?),
 	    "creation-date" => Parameter::CreationDate(
                 Parameter::parse_creation(value)
@@ -726,11 +724,11 @@ async fn create_key(config: &mut crate::Config<'_>, filename: &str, i: usize,
     let mut agent = config.connect_agent().await?;
     let passphrase_store;
     if passphrase.is_none() && ! no_protection {
-        passphrase_store = crate::agent::get_passphrase(
+        passphrase_store = config.get_passphrase(
             &mut agent, &None, &None,
             Some("Passphrase:".into()),
             Some("Please enter passphrase to protect your new key".into()),
-            false, 1, true, true, |_, _| None).await?;
+            false, 1, true, true, |_| Ok(())).await?;
         passphrase = Some(&passphrase_store);
     };
 
@@ -848,11 +846,9 @@ async fn create_key(config: &mut crate::Config<'_>, filename: &str, i: usize,
 
     // Store the secrets in the agent.
     for subkey in cert.keys().secret() {
-        // See if we import a new key or subkey.
-        crate::agent::import(&mut agent,
-                             config.policy(),
-                             &cert, &subkey,
-                             config.batch).await?;
+        agent.import(config.policy(),
+                     &cert, &subkey,
+                     config.batch, false).await?;
     }
 
     // Set to ultimately trusted.
