@@ -1,5 +1,6 @@
 use std::{
     cell::{OnceCell, RefCell},
+    collections::BTreeSet,
     fmt,
     fs,
     io::{self, Read, Write},
@@ -14,6 +15,8 @@ use indexmap::IndexMap;
 use sequoia_openpgp as openpgp;
 use sequoia_ipc as ipc;
 use openpgp::{
+    Fingerprint,
+    KeyHandle,
     cert::prelude::*,
     crypto::Password,
     packet::{
@@ -30,19 +33,22 @@ use sequoia_cert_store::{
     Store,
 };
 
+use sequoia_gpg_agent as gpg_agent;
+
 pub mod gnupg_interface;
 
+#[macro_use]
+mod print;
 #[macro_use]
 mod macros;
 pub mod tracing;
-pub mod agent;
 #[macro_use]
 pub mod argparse;
 use argparse::{Argument, Opt, flags::*};
 pub mod babel;
 pub mod clock;
 pub mod common;
-use common::{Common, Compliance, Query, Validity};
+use common::{Common, Compliance, Query, TrustModel, Validity};
 pub mod compliance;
 pub mod homedir;
 mod interactive;
@@ -71,6 +77,9 @@ pub mod parcimonie;
 pub mod dirmngr;
 pub mod migrate;
 pub mod generate_key;
+pub mod filter;
+pub mod quick;
+pub mod assert_pubkey_algo;
 
 /// Commands and options.
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
@@ -436,6 +445,7 @@ pub enum CmdOrOpt {
     oForceSignKey,
     oForbidGenKey,
     oRequireCompliance,
+    oAssertPubkeyAlgo,
 
     oNoop,
 
@@ -443,6 +453,7 @@ pub enum CmdOrOpt {
     aXSequoiaParcimonie,
     oXSequoiaAutostartParcimonie,
     aXSequoiaParcimonieDaemonize,
+    oXSequoiaQuietFakedSystemTime,
 
     // Special, implicit commands.
     aHelp = 'h' as isize,
@@ -472,6 +483,9 @@ pub struct Config<'store> {
     trust_model_impl: Box<dyn trust::Model>,
     de_vs_producer: compliance::DeVSProducer,
 
+    /// Emulates GnuPG's pk_cache.
+    pk_cache: Mutex<BTreeSet<Fingerprint>>,
+
     // Configuration.
     answer_no: bool,
     answer_yes: bool,
@@ -481,19 +495,19 @@ pub struct Config<'store> {
     ask_sig_expire: bool,
     auto_key_locate: Vec<AutoKeyLocate>,
     batch: bool,
-    cert_digest: HashAlgorithm,
+    cert_digest: Option<HashAlgorithm>,
     cert_policy_url: Vec<URL>,
     check_sigs: bool,
     comments: Vec<String>,
     completes_needed: Option<i64>,
     compliance: Compliance,
-    compress_algo: CompressionAlgorithm,
+    compress_algo: Option<CompressionAlgorithm>,
     compress_level: i64,
     debug: u32,
     def_cert_expire: Option<time::Duration>,
     def_cert_level: i64,
-    def_cipher: SymmetricAlgorithm,
-    def_digest: HashAlgorithm,
+    def_cipher: Option<SymmetricAlgorithm>,
+    def_digest: Option<HashAlgorithm>,
     def_keyserver_url: Option<KeyserverURL>,
     def_preferences: Preferences,
     def_recipient: Option<String>,
@@ -518,6 +532,7 @@ pub struct Config<'store> {
     input_size_hint: Option<u64>,
     interactive: bool,
     keydb: keydb::KeyDB<'store>,
+    keyid_format: KeyIDFormat,
     keyserver: Vec<KeyserverURL>,
     keyserver_options: keyserver::KeyserverOptions,
     list_only: bool,
@@ -537,11 +552,11 @@ pub struct Config<'store> {
     outfile: Option<String>,
     override_session_key: Option<SessionKey>,
     passphrase_repeat: i64,
-    personal_cipher_prefs: Vec<SymmetricAlgorithm>,
-    personal_digest_prefs: Vec<HashAlgorithm>,
-    personal_compress_prefs: Vec<CompressionAlgorithm>,
+    personal_cipher_prefs: Option<Vec<SymmetricAlgorithm>>,
+    personal_digest_prefs: Option<Vec<HashAlgorithm>>,
+    personal_compress_prefs: Option<Vec<CompressionAlgorithm>>,
     photo_viewer: Option<PathBuf>,
-    pinentry_mode: agent::PinentryMode,
+    pinentry_mode: gpg_agent::PinentryMode,
     quiet: bool,
     remote_user: Vec<Recipient>,
     request_origin: RequestOrigin,
@@ -577,6 +592,9 @@ pub struct Config<'store> {
     with_tofu_info: bool,
     with_wkd_hash: bool,
 
+    // Backported from GnuPG 2.4.5.
+    pubkey_algo_policy: assert_pubkey_algo::Policy,
+
     // Extension.
     autostart_parcimonie: bool,
 
@@ -599,6 +617,9 @@ impl<'store> Config<'store> {
             trust_model_impl: common::null_model(),
             de_vs_producer: compliance::DeVSProducer::default(),
 
+            // Emulation.
+            pk_cache: Default::default(),
+
             // Configuration.
             answer_no: false,
             answer_yes: false,
@@ -614,7 +635,7 @@ impl<'store> Config<'store> {
             comments: vec![],
             completes_needed: None,
             compliance: Default::default(),
-            compress_algo: CompressionAlgorithm::Uncompressed,
+            compress_algo: Default::default(),
             compress_level: 5,
             debug: 0,
             def_cert_expire: None,
@@ -648,6 +669,7 @@ impl<'store> Config<'store> {
             input_size_hint: None,
             interactive: false,
             keydb: keydb::KeyDB::for_gpg(),
+            keyid_format: Default::default(),
             keyserver: Default::default(),
             keyserver_options: Default::default(),
             list_only: false,
@@ -667,9 +689,9 @@ impl<'store> Config<'store> {
             outfile: None,
             override_session_key: None,
             passphrase_repeat: 0, // XXX
-            personal_cipher_prefs: Preferences::default().symmetric,
-            personal_digest_prefs: Preferences::default().hash,
-            personal_compress_prefs: Preferences::default().compression,
+            personal_cipher_prefs: None,
+            personal_digest_prefs: None,
+            personal_compress_prefs: None,
             photo_viewer: None,
             pinentry_mode: Default::default(),
             quiet: false,
@@ -707,12 +729,15 @@ impl<'store> Config<'store> {
             with_tofu_info: false,
             with_wkd_hash: false,
 
+            // Backported from GnuPG 2.4.5.
+            pubkey_algo_policy: Default::default(),
+
             // Extensions.
             autostart_parcimonie: false,
 
             // Streams.
             attribute_fd: Box::new(io::sink()),
-            command_fd: io::stdin().into(),
+            command_fd: interactive::Fd::interactive(),
             logger_fd: Mutex::new(RefCell::new(Box::new(io::stderr()))),
             status_fd: status::Fd::sink(),
         })
@@ -720,7 +745,7 @@ impl<'store> Config<'store> {
 
     /// Emits the usage and terminates the process.
     pub fn wrong_args(&self, msg: fmt::Arguments) -> ! {
-        eprintln!("usage: gpg [options] {}", msg);
+        safe_eprintln!("usage: gpg [options] {}", msg);
         std::process::exit(2);
     }
 
@@ -730,17 +755,14 @@ impl<'store> Config<'store> {
     }
 
     /// Returns a connection to the GnuPG agent.
-    pub async fn connect_agent(&self) -> Result<ipc::gnupg::Agent> {
-        use agent::{connect, send_simple};
+    pub async fn connect_agent(&self) -> Result<gpg_agent::Agent> {
+        let mut agent = gpg_agent::Agent::connect_to(&self.homedir).await?;
 
-        let ctx = self.ipc()?;
-        let mut agent = connect(ctx).await?;
-
-        send_simple(&mut agent, "RESET").await?;
+        agent.send_simple("RESET").await?;
 
         if let Ok(tty) = std::env::var("GPG_TTY") {
-            send_simple(&mut agent, format!(
-                "OPTION ttyname={}", tty)).await?;
+            agent.send_simple(
+                format!("OPTION ttyname={}", tty)).await?;
         } else {
             #[cfg(unix)]
             {
@@ -755,36 +777,36 @@ impl<'store> Config<'store> {
                 };
 
                 if let Some(tty) = tty {
-                    send_simple(&mut agent, format!(
-                        "OPTION ttyname={}", tty)).await?;
+                    agent.send_simple(
+                        format!("OPTION ttyname={}", tty)).await?;
                 }
             }
         }
         let ttyname = unsafe { libc::ttyname(0) };
         if ! ttyname.is_null() {
             let ttyname = unsafe { std::ffi::CStr::from_ptr(ttyname) };
-            send_simple(&mut agent, format!(
+            agent.send_simple(format!(
                 "OPTION ttyname={}",
                 String::from_utf8_lossy(ttyname.to_bytes()))).await?;
         }
         if let Ok(term) = std::env::var("TERM") {
-            send_simple(&mut agent, format!("OPTION ttytype={}", term)).await?;
+            agent.send_simple(format!("OPTION ttytype={}", term)).await?;
         }
         if let Ok(display) = std::env::var("DISPLAY") {
-            send_simple(&mut agent, format!("OPTION display={}", display)).await?;
+            agent.send_simple(format!("OPTION display={}", display)).await?;
         }
         if let Ok(xauthority) = std::env::var("XAUTHORITY") {
-            send_simple(&mut agent, format!("OPTION xauthority={}", xauthority)).await?;
+            agent.send_simple(format!("OPTION xauthority={}", xauthority)).await?;
         }
         if let Ok(dbus) = std::env::var("DBUS_SESSION_BUS_ADDRESS") {
-            send_simple(&mut agent,
-                        format!("OPTION putenv=DBUS_SESSION_BUS_ADDRESS={}",
-                                dbus)).await?;
+            agent.send_simple(
+                format!("OPTION putenv=DBUS_SESSION_BUS_ADDRESS={}",
+                        dbus)).await?;
         }
-        send_simple(&mut agent, "OPTION allow-pinentry-notify").await?;
-        send_simple(&mut agent, "OPTION agent-awareness=2.1.0").await?;
-        send_simple(&mut agent, format!("OPTION pinentry-mode={}",
-                                        self.pinentry_mode.as_str())).await?;
+        agent.send_simple("OPTION allow-pinentry-notify").await?;
+        agent.send_simple("OPTION agent-awareness=2.1.0").await?;
+        agent.send_simple(format!("OPTION pinentry-mode={}",
+                                  self.pinentry_mode.as_str())).await?;
 
         Ok(agent)
     }
@@ -817,17 +839,17 @@ impl<'store> Config<'store> {
                 let m = std::fs::metadata(&self.homedir)?;
 
                 if ! m.is_dir() {
-                    eprintln!("gpg: WARNING: homedir '{}' is not a directory",
+                    safe_eprintln!("gpg: WARNING: homedir '{}' is not a directory",
                               self.homedir.display());
                 }
 
                 if m.uid() != unsafe { libc::getuid() } {
-                    eprintln!("gpg: WARNING: unsafe ownership on homedir '{}'",
+                    safe_eprintln!("gpg: WARNING: unsafe ownership on homedir '{}'",
                               self.homedir.display());
                 }
 
                 if m.mode() & (libc::S_IRWXG | libc::S_IRWXO) as u32 > 0 {
-                    eprintln!("gpg: WARNING: unsafe permissions on homedir '{}'",
+                    safe_eprintln!("gpg: WARNING: unsafe permissions on homedir '{}'",
                               self.homedir.display());
                 }
             },
@@ -851,14 +873,12 @@ impl<'store> Config<'store> {
                             -> Result<Box<dyn openpgp::crypto::Signer + Send + Sync>>
     {
         let mut agent = self.connect_agent().await?;
-        agent::has_key(&mut agent, subkey).await?;
+        agent.has_key(subkey).await?;
 
-        let ctx = self.ipc()?;
-        let mut pair = ipc::gnupg::KeyPair::new(&ctx, subkey)?
-            .with_cert(vcert);
+        let mut pair = agent.keypair(subkey)?.with_cert(vcert);
 
         // See if we have a static password to loop back to the agent.
-        if let (agent::PinentryMode::Loopback, Some(p)) =
+        if let (gpg_agent::PinentryMode::Loopback, Some(p)) =
             (&self.pinentry_mode, self.static_passphrase.borrow().as_ref())
         {
             pair = pair.with_password(p.clone());
@@ -867,6 +887,61 @@ impl<'store> Config<'store> {
         Ok(Box::new(pair))
     }
 
+    /// Makes the agent ask for a password.
+    pub async fn get_passphrase<P>(&self,
+                                   agent: &mut gpg_agent::Agent,
+                                   cache_id: &Option<String>,
+                                   err_msg: &Option<String>,
+                                   prompt: Option<String>,
+                                   desc_msg: Option<String>,
+                                   newsymkey: bool,
+                                   repeat: usize,
+                                   check: bool,
+                                   qualitybar: bool,
+                                   mut pinentry_launched_cb: P)
+                                   -> Result<Password>
+    where
+        P: FnMut(&[u8]) -> Result<()>,
+    {
+        use gpg_agent::PinentryMode;
+        use gpg_agent::sequoia_ipc::assuan::Response;
+
+        let callback = |_agent: &mut _, response| {
+            if let Response::Inquire { keyword, parameters } = &response {
+                match (keyword.as_str(), parameters, &self.pinentry_mode) {
+                    ("PASSPHRASE", _, PinentryMode::Loopback) => {
+                        // Do we have a pre-set password?
+                        if let Some(p) =
+                            self.static_passphrase.borrow().as_ref()
+                        {
+                            return Ok(Some(p.map(|decrypted| decrypted.clone())));
+                        }
+
+                        // We don't.  Prompt for a password.
+                        let _ = // XXX.
+                            self.status().emit(status::Status::InquireMaxLen(100));
+                        let p = self.prompt_password()?;
+                        Ok(Some(p.map(|decrypted| decrypted.clone())))
+                    },
+                    ("PINENTRY_LAUNCHED", Some(p), _) => {
+                        let _ = // XXX.
+                            pinentry_launched_cb(p.as_slice());
+                        Ok(None)
+                    },
+                    // We silently ignore unknown inquiries.
+                    // Alternatively, we could return an error.
+                    _ => Ok(None),
+                }
+            } else {
+                Ok(None)
+            }
+        };
+
+        Ok(agent.get_passphrase(cache_id, err_msg, prompt, desc_msg,
+                                newsymkey, repeat, check, qualitybar,
+                                callback).await?)
+    }
+
     /// Returns the local users used e.g. in signing operations.
     pub async fn local_users(&self, flags: KeyFlags) -> Result<Vec<String>> {
         if self.local_user.is_empty() {
@@ -900,7 +975,7 @@ impl<'store> Config<'store> {
                         for sk in vcert.keys().key_flags(&flags).alive()
                             .revoked(false)
                         {
-                            if agent::has_key(&mut agent, sk.key()).await? {
+                            if agent.has_key(sk.key()).await? {
                                 return Ok(vec![cert.fingerprint().to_string()]);
                             }
                         }
@@ -929,21 +1004,9 @@ impl<'store> Config<'store> {
     /// Returns certs matching a given query using groups and the
     /// configured trust model.
     pub fn lookup_certs(&self, query: &Query) -> Result<Vec<(Validity, Arc<LazyCert<'store>>)>> {
-        let certs = self.lookup_certs_with(
+        self.lookup_certs_with(
             self.trust_model_impl.with_policy(self, Some(self.now()))?.as_ref(),
-            query, true)?;
-
-        // GnuPG emits a key considered status message on every
-        // lookup, even if it is repeated later on.  Do the same.
-        for (_, cert) in &certs {
-            self.status().emit(
-                status::Status::KeyConsidered {
-                    fingerprint: cert.fingerprint(),
-                    not_selected: false,
-                    all_expired_or_revoked: false,
-                })?;
-        }
-        Ok(certs)
+            query, true)
     }
 
     /// Returns certs matching a given query using groups and the
@@ -965,7 +1028,7 @@ impl<'store> Config<'store> {
                 if let Some(queries) = self.groups.get(e.as_str()) {
                     let mut acc = Vec::new();
                     for query in queries {
-                        let q = query.as_str().into();
+                        let q = query.parse()?;
                         acc.append(
                             &mut self.lookup_certs_with(vtm, &q, false)?);
                     }
@@ -977,7 +1040,7 @@ impl<'store> Config<'store> {
                 if let Some(queries) = self.groups.get(&f[..]) {
                     let mut acc = Vec::new();
                     for query in queries {
-                        let q = query.as_str().into();
+                        let q = query.parse()?;
                         acc.append(
                             &mut self.lookup_certs_with(vtm, &q, false)?);
                     }
@@ -990,6 +1053,34 @@ impl<'store> Config<'store> {
         vtm.lookup(query)
     }
 
+    /// Like `.keydb().lookup_by_cert_or_subkey` but emits
+    /// KEY_CONSIDERED lines.
+    pub fn lookup_by_cert_or_subkey(&self, kh: &KeyHandle)
+                                    -> Result<Vec<Arc<LazyCert<'store>>>>
+    {
+        self.keydb().lookup_by_cert_or_subkey(kh)
+            .map(|certs| {
+                for cert in &certs {
+                    let fp = cert.fingerprint();
+                    {
+                        let mut pk_cache = self.pk_cache.lock().unwrap();
+                        if pk_cache.contains(&fp) {
+                            continue;
+                        } else {
+                            pk_cache.insert(fp);
+                        }
+                    }
+
+                    let _ = self.status().emit(status::Status::KeyConsidered {
+                        fingerprint: cert.fingerprint(),
+                        not_selected: false,
+                        all_expired_or_revoked: false,
+                    });
+                }
+                certs
+            })
+    }
+
     /// Stores a revocation certificate.
     pub fn store_revocation(&self, cert: &Cert, rev: Signature) -> Result<()> {
         let store = self.homedir().join("openpgp-revocs.d");
@@ -1063,8 +1154,7 @@ impl<'store> common::Common<'store> for Config<'store> {
         self.log(format_args!("{}: {}", self.argv0(), msg));
     }
 
-    fn error(&self, msg: fmt::Arguments) {
-        self.warn(msg);
+    fn fail(&self) {
         self.fail.set(true);
     }
 
@@ -1132,7 +1222,6 @@ struct Flags {
     disable_signer_uid: bool,
     force_sign_key: bool,
     include_key_block: bool,
-    use_embedded_filename: bool,
 }
 
 #[derive(Clone)]
@@ -1206,6 +1295,34 @@ impl std::str::FromStr for RequestOrigin {
     }
 }
 
+#[derive(Clone)]
+enum KeyIDFormat {
+    None,
+    Long,
+    HexLong,
+}
+
+impl Default for KeyIDFormat {
+    fn default() -> Self {
+        KeyIDFormat::None
+    }
+}
+
+impl std::str::FromStr for KeyIDFormat {
+    type Err = anyhow::Error;
+
+    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
+        match s.to_lowercase().as_str() {
+            "none" => Ok(KeyIDFormat::None),
+            "short" | "0xshort" => Err(anyhow::anyhow!(
+                "short key IDs are not supported by the Sequoia Chameleon")),
+            "long" => Ok(KeyIDFormat::Long),
+            "0xlong" => Ok(KeyIDFormat::HexLong),
+            _ => Err(anyhow::anyhow!("invalid key ID format {:?}", s)),
+        }
+    }
+}
+
 #[derive(Debug, Clone)]
 pub struct Preferences {
     hash: Vec<HashAlgorithm>,
@@ -1308,11 +1425,11 @@ fn set_cmd(cmd: &mut Option<CmdOrOpt>, new_cmd: CmdOrOpt)
 }
 
 fn obsolete_option(s: &str) {
-    eprintln!("WARNING: {:?} is an obsolete option - it has no effect", s);
+    safe_eprintln!("WARNING: {:?} is an obsolete option - it has no effect", s);
 }
 
 fn deprecated_warning(s: &str, repl1: &str, repl2: &str) {
-    eprintln!("WARNING: {:?} is a deprecated option, \
+    safe_eprintln!("WARNING: {:?} is a deprecated option, \
                please use \"{}{}\" instead",
               s, repl1, repl2);
 }
@@ -1414,10 +1531,10 @@ impl std::str::FromStr for SessionKey {
 }
 
 fn print_additional_version(config: &Config) {
-    println!();
-    println!("Home: {}", config.homedir.display());
+    safe_println!();
+    safe_println!("Home: {}", config.homedir.display());
 
-    println!("Supported algorithms:");
+    safe_println!("Supported algorithms:");
 
     struct Writer(String, usize);
     impl Writer {
@@ -1444,7 +1561,7 @@ fn print_additional_version(config: &Config) {
                 Some(c) => self.0.push(c), // Put back.  Unlikely.
                 None => (),
             }
-            println!("{}", self.0);
+            safe_println!("{}", self.0);
             self.0.clear();
         }
     }
@@ -1506,6 +1623,7 @@ fn real_main() -> anyhow::Result<()> {
 
     let parser = argparse::Parser::new(
         "gpg",
+        crate::gnupg_interface::VERSION,
         "Sign, check, encrypt or decrypt\n\
          Default operation depends on the input data",
         &OPTIONS)
@@ -1673,6 +1791,8 @@ fn real_main() -> anyhow::Result<()> {
 	    aShowKeys =>
             {
                 set_cmd(&mut command, cmd)?;
+                // Note: contrary to GnuPG we don't use the import
+                // code.
                 opt.import_options.show = true;
                 opt.import_options.dry_run = true;
                 opt.list_options.unusable_uids = true;
@@ -1830,14 +1950,14 @@ fn real_main() -> anyhow::Result<()> {
 
 	    oStatusFD => {
                 opt.status_fd =
-                    utils::sink_from_fd(value.as_int().unwrap())?.into();
+                    argparse::utils::sink_from_fd(value.as_int().unwrap())?.into();
             },
 	    oStatusFile => {
                 opt.status_fd =
                     Box::new(fs::File::create(value.as_str().unwrap())?).into();
             },
 	    oAttributeFD => {
-                opt.attribute_fd = utils::sink_from_fd(value.as_int().unwrap())?;
+                opt.attribute_fd = argparse::utils::sink_from_fd(value.as_int().unwrap())?;
             },
 	    oAttributeFile => {
                 opt.attribute_fd =
@@ -1845,7 +1965,7 @@ fn real_main() -> anyhow::Result<()> {
             },
 	    oLoggerFD => {
                 opt.logger_fd = Mutex::new(RefCell::new(
-                    utils::sink_from_fd(value.as_int().unwrap())?));
+                    argparse::utils::sink_from_fd(value.as_int().unwrap())?));
             },
             oLoggerFile => {
                 // XXX: Why is this different from opt.logger_fd??
@@ -1853,8 +1973,9 @@ fn real_main() -> anyhow::Result<()> {
             },
 
 	    oWithFingerprint => {
-                opt.with_fingerprint = true;
                 opt.fingerprint += 1;
+                opt.with_fingerprint = true;
+                opt.with_subkey_fingerprint = opt.fingerprint > 1;
             },
 	    oWithSubkeyFingerprint => {
                 opt.with_subkey_fingerprint = true;
@@ -2010,7 +2131,7 @@ fn real_main() -> anyhow::Result<()> {
 	    },
 
 	    oForceOwnertrust => {
-	        eprintln!("Note: {} is not for normal use!",
+	        safe_eprintln!("Note: {} is not for normal use!",
 		          "--force-ownertrust");
 	        opt.force_ownertrust = value.as_str().unwrap().parse()?;
 	    },
@@ -2084,12 +2205,13 @@ fn real_main() -> anyhow::Result<()> {
 	    oSigKeyserverURL => {
                 opt.sig_keyserver_url.push(URL::new(value.as_str().unwrap()));
             },
-	    oUseEmbeddedFilename => {
-	        opt.flags.use_embedded_filename = true;
-	    },
-	    oNoUseEmbeddedFilename => {
-	        opt.flags.use_embedded_filename = false;
-	    },
+
+	    oUseEmbeddedFilename =>
+                return Err(anyhow::anyhow!(
+                    "This option is a security risk \
+                     and is thus not supported")),
+	    oNoUseEmbeddedFilename => (), // This is a NOP for us.
+
 	    oComment => {
 	        if let Ok(v) = value.as_str() {
 	            opt.comments.push(v.into());
@@ -2213,7 +2335,7 @@ fn real_main() -> anyhow::Result<()> {
             },
 	    oDefSigExpire => {
 		opt.def_sig_expire = utils::parse_expiration(
-                    &opt, value.as_str().unwrap())?;
+                    opt.now(), value.as_str().unwrap())?;
 	    },
 	    oAskSigExpire => {
                 opt.ask_sig_expire = true;
@@ -2223,7 +2345,7 @@ fn real_main() -> anyhow::Result<()> {
             },
 	    oDefCertExpire => {
 		opt.def_cert_expire = utils::parse_expiration(
-                    &opt, value.as_str().unwrap())?;
+                    opt.now(), value.as_str().unwrap())?;
 	    },
 	    oAskCertExpire => {
                 opt.ask_cert_expire = true;
@@ -2271,7 +2393,7 @@ fn real_main() -> anyhow::Result<()> {
                     Some(value.as_str().unwrap().into()).into();
 	    },
 	    oPassphraseFD => {
-                pwfd = Some(utils::source_from_fd(value.as_int().unwrap())?);
+                pwfd = Some(argparse::utils::source_from_fd(value.as_int().unwrap())?);
             },
 	    oPassphraseFile => {
                 pwfd = Some(fs::File::open(value.as_str().unwrap())?);
@@ -2290,7 +2412,7 @@ fn real_main() -> anyhow::Result<()> {
 
 	    oCommandFD => {
                 opt.command_fd =
-                    utils::source_from_fd(value.as_int().unwrap())?.into();
+                    argparse::utils::source_from_fd(value.as_int().unwrap())?.into();
             },
 	    oCommandFile => {
                 opt.command_fd =
@@ -2298,13 +2420,13 @@ fn real_main() -> anyhow::Result<()> {
             },
 
 	    oCipherAlgo => opt.def_cipher =
-                value.as_str().unwrap().parse::<babel::Fish<_>>()?.0,
+                Some(value.as_str().unwrap().parse::<babel::Fish<_>>()?.0),
 	    oDigestAlgo => opt.def_digest =
-                value.as_str().unwrap().parse::<babel::Fish<_>>()?.0,
+                Some(value.as_str().unwrap().parse::<babel::Fish<_>>()?.0),
 	    oCompressAlgo => opt.compress_algo =
-                value.as_str().unwrap().parse::<babel::Fish<_>>()?.0,
+                Some(value.as_str().unwrap().parse::<babel::Fish<_>>()?.0),
 	    oCertDigestAlgo => opt.cert_digest =
-                value.as_str().unwrap().parse::<babel::Fish<_>>()?.0,
+                Some(value.as_str().unwrap().parse::<babel::Fish<_>>()?.0),
 
 	    oNoSecmemWarn => (),
 	    oRequireSecmem => (),
@@ -2385,15 +2507,18 @@ fn real_main() -> anyhow::Result<()> {
                     Some(value.as_str().unwrap().parse()?);
             },
             oOverrideSessionKeyFD => {
-                let mut h = utils::source_from_fd(value.as_int().unwrap())?;
+                let mut h = argparse::utils::source_from_fd(value.as_int().unwrap())?;
                 let mut buf = Vec::new();
                 h.read_to_end(&mut buf)?;
                 opt.override_session_key =
                     Some(String::from_utf8(buf)?.parse()?);
             },
             oTrustedKey => {
-                // XXX: We don't really support KeyIDs here.
-                opt.trusted_keys.push(value.as_str().unwrap().parse()?);
+                // XXX: We don't support KeyIDs here.
+                let fp = value.as_str().unwrap().parse()?;
+                if let Err(i) = opt.trusted_keys.binary_search(&fp) {
+                    opt.trusted_keys.insert(i, fp);
+                }
             },
 	    oFastListMode => opt.list_options.fast_list = true,
 	    oFixedListMode => (), // This is a NOP in GnuPG.
@@ -2408,15 +2533,15 @@ fn real_main() -> anyhow::Result<()> {
                 opt.def_keyserver_url = Some(value.as_str().unwrap().parse()?),
             oPersonalCipherPreferences =>
                 if let Some(p) = Preferences::parse(value.as_str().unwrap())? {
-                    opt.personal_cipher_prefs = p.symmetric;
+                    opt.personal_cipher_prefs = Some(p.symmetric);
                 },
             oPersonalDigestPreferences =>
                 if let Some(p) = Preferences::parse(value.as_str().unwrap())? {
-                    opt.personal_digest_prefs = p.hash;
+                    opt.personal_digest_prefs = Some(p.hash);
                 },
             oPersonalCompressPreferences =>
                 if let Some(p) = Preferences::parse(value.as_str().unwrap())? {
-                    opt.personal_compress_prefs = p.compression;
+                    opt.personal_compress_prefs = Some(p.compression);
                 },
             oWeakDigest => {
                 opt.policy.weak_digest(
@@ -2437,7 +2562,7 @@ fn real_main() -> anyhow::Result<()> {
                 }
             },
             oUnGroup => {
-                opt.groups.remove(value.as_str().unwrap());
+                opt.groups.shift_remove(value.as_str().unwrap());
             },
             oNoGroups => {
                 opt.groups.clear();
@@ -2445,6 +2570,8 @@ fn real_main() -> anyhow::Result<()> {
             oMultifile => {
                 multifile = true;
             },
+            oKeyidFormat =>
+                opt.keyid_format = value.as_str().unwrap().parse()?,
             oAutoKeyLocate => {
                 auto_key_locate_given = true;
                 for s in value.as_str().unwrap().split(',') {
@@ -2463,19 +2590,27 @@ fn real_main() -> anyhow::Result<()> {
                 auto_key_locate_given = true;
                 opt.auto_key_locate.clear();
             },
-            oFakedSystemTime => {
+
+            oFakedSystemTime | oXSequoiaQuietFakedSystemTime => {
                 opt.clock = value.as_str().unwrap().parse()?;
-                // XXX: GnuPG prints this warning later.
-                use chrono::{DateTime, Utc};
-                opt.warn(format_args!(
-                    "WARNING: running with faked system time: {}",
-                    // 2022-09-19 10:37:42
-                    DateTime::<Utc>::from(opt.now())
-                        .format("%Y-%m-%d %H:%M:%S")));
+                if cmd == oFakedSystemTime {
+                    // XXX: GnuPG prints this warning later.
+                    use chrono::{DateTime, Utc};
+                    opt.warn(format_args!(
+                        "WARNING: running with faked system time: {}",
+                        // 2022-09-19 10:37:42
+                        DateTime::<Utc>::from(opt.now())
+                            .format("%Y-%m-%d %H:%M:%S")));
+                }
             },
 
             oForbidGenKey => opt.forbid_gen_key = true,
 
+            // Backported from GnuPG 2.4.5.
+            oAssertPubkeyAlgo =>
+                opt.pubkey_algo_policy.handle_cmdline_arg(
+                    value.as_str().unwrap())?,
+
             // Our own extensions.
             aXSequoiaParcimonie => {
                 set_cmd(&mut command, aXSequoiaParcimonie)?;
@@ -2517,15 +2652,15 @@ fn real_main() -> anyhow::Result<()> {
     }
 
     if greeting && ! no_greeting {
-        eprintln!("gpg (Sequoia Chameleon {}) {}; \
+        safe_eprintln!("gpg (Sequoia Chameleon {}) {}; \
                    Copyright (C) 2024 Sequoia PGP",
                   env!("CARGO_PKG_VERSION"),
                   crate::gnupg_interface::VERSION);
-        eprintln!("This is free software: \
+        safe_eprintln!("This is free software: \
                    you are free to change and redistribute it.");
-        eprintln!("There is NO WARRANTY, \
+        safe_eprintln!("There is NO WARRANTY, \
                    to the extent permitted by law.");
-        eprintln!();
+        safe_eprintln!();
     }
 
     if multifile {
@@ -2567,6 +2702,11 @@ fn real_main() -> anyhow::Result<()> {
     // Get the default one if no keyring has been specified.
     if keyrings.is_empty() {
         opt.keydb.add_resource(&opt.homedir, "pubring.gpg", false, true)?;
+
+        // Use the keybox database if it is there.
+        let _ = opt.keydb.add_resource(
+            &opt.homedir, "gnupg-kbx-db:public-keys.d/pubring.db",
+            true, false)?;
     }
 
     for path in keyrings {
@@ -2597,7 +2737,7 @@ fn real_main() -> anyhow::Result<()> {
         || (matches!(command, Some(aExport)) && args.len() == 0);
     opt.keydb.initialize(! preload)?;
     opt.trust_model_impl =
-        opt.trust_model.unwrap_or_default().build(&opt)?;
+        opt.trust_model.unwrap_or(TrustModel::Auto).build(&opt)?;
     opt.trustdb.read_ownertrust(opt.trustdb.path(&opt))?;
 
     // Read the owner-trusts from our DB.
@@ -2614,7 +2754,7 @@ fn real_main() -> anyhow::Result<()> {
         opt.quiet = quiet;
     }
 
-    if let agent::PinentryMode::Loopback = opt.pinentry_mode {
+    if let gpg_agent::PinentryMode::Loopback = opt.pinentry_mode {
         // In loopback mode, never ask for the password multiple
         // times.
 	opt.passphrase_repeat = 0;
@@ -2730,13 +2870,20 @@ fn real_main() -> anyhow::Result<()> {
             opt.list_options.list_sigs = true;
             list_keys::cmd_list_keys(&mut opt, &args, false)
         },
+        Some(aCheckKeys) => {
+            opt.check_sigs = true;
+            opt.list_options.list_sigs = true;
+            list_keys::cmd_list_keys(&mut opt, &args, false)
+        },
         Some(aListSecretKeys) =>
             list_keys::cmd_list_keys(&mut opt, &args, true),
+        Some(aShowKeys) =>
+            list_keys::cmd_show_keys(&mut opt, &args),
         Some(aCheckTrustDB) => Ok(()), // This is a NOP for us.
         Some(aImportOwnerTrust) =>
-            trust::db::cmd_import_ownertrust(&mut opt, &args),
+            commands::cmd_import_ownertrust(&mut opt, &args),
         Some(aExportOwnerTrust) =>
-            trust::db::cmd_export_ownertrust(&opt, &args),
+            commands::cmd_export_ownertrust(&opt, &args),
         Some(aListConfig) => commands::cmd_list_config(&opt, &args),
         Some(aGenRevoke) => commands::cmd_generate_revocation(&opt, &args),
         Some(aEnArmor) => commands::cmd_enarmor(&opt, &args),
@@ -2749,6 +2896,8 @@ fn real_main() -> anyhow::Result<()> {
         Some(aKeygen) => generate_key::cmd_generate_key(&mut opt, &args, false),
         Some(aFullKeygen) => generate_key::cmd_generate_key(&mut opt, &args, true),
         Some(aQuickAddKey) => generate_key::cmd_quick_add_key(&mut opt, &args),
+        Some(aQuickAddUid) => quick::cmd_quick_add_uid(&mut opt, &args),
+        Some(aQuickRevUid) => quick::cmd_quick_revoke_uid(&mut opt, &args),
         None => commands::cmd_implicit(&opt, &args),
 
         // Our own extensions.
@@ -2797,7 +2946,7 @@ fn real_main() -> anyhow::Result<()> {
             if opt.verbose > 1 {
                 print_error_chain(&e);
             } else {
-                eprintln!("gpg: {}", e);
+                safe_eprintln!("gpg: {}", e);
             }
             std::process::exit(2);
         }
