@@ -70,7 +70,7 @@ pub struct ImportOptions {
 
 impl ImportOptions {
     const OPTS: [Opt<ImportOptions>; 20] = [
-        opt_todo! {
+        opt! {
             "import-local-sigs",
             |o, s, _| Ok({ o.local_sigs = s; }),
             "import signatures that are marked as local-only",
@@ -137,7 +137,7 @@ impl ImportOptions {
         },
 
         /* Aliases for backward compatibility */
-        opt_todo! {
+        opt! {
             "allow-local-sigs",
             |o, s, _| Ok({ o.local_sigs = s; }),
             "",
@@ -258,6 +258,19 @@ pub async fn do_import_cert(config: &mut crate::Config<'_>,
     use crate::status::*;
     let mut flags = crate::status::ImportOkFlags::default();
 
+    // Maybe strip out local signatures.
+    let cert = if config.import_options.local_sigs {
+        cert
+    } else {
+        let mut p = cert.into_tsk().into_packets().collect::<Vec<_>>();
+        p.retain(|p| match p {
+            Packet::Signature(s) =>
+                s.exportable_certification().unwrap_or(true),
+            _ => true,
+        });
+        Cert::from_packets(p.into_iter())?
+    };
+
     // We import the cert first, if this is a key, we'll deal
     // with the secrets later.
     let (cert, key) =
@@ -310,7 +323,7 @@ pub async fn do_import_cert(config: &mut crate::Config<'_>,
                 list_keys::async_list_keys(
                     config,
                     vec![key.clone()].into_iter(),
-                    true, false, false, false,
+                    true, false, false, false, false,
                     std::io::stdout()).await?;
             } else if for_migration {
                 // Be quiet when migrating.
@@ -420,7 +433,7 @@ pub async fn do_import_cert(config: &mut crate::Config<'_>,
         if config.import_options.show {
             list_keys::async_list_keys(
                 config, vec![key.clone()].into_iter(),
-                true, false, false, false,
+                true, false, false, false, false,
                 std::io::stdout()).await?;
         } else if for_migration {
             // Be quiet when migrating.
@@ -466,13 +479,20 @@ pub async fn do_import_cert(config: &mut crate::Config<'_>,
 
         for subkey in key.to_cert()?.keys().secret() {
             // See if we import a new key or subkey.
-            let c = crate::agent::import(&mut agent,
-                                         config.policy(),
-                                         key.to_cert()?, &subkey,
-                                         config.batch).await?;
-
-            changed |= c;
-            unchanged |= !c;
+            match agent.import(config.policy(),
+                               key.to_cert()?, &subkey,
+                               config.batch, false).await
+            {
+                Ok(_) => changed |= true,
+                Err(e) => {
+                    use sequoia_gpg_agent::Error::KeyExists;
+                    if let Some(KeyExists(_, _)) =  e.downcast_ref() {
+                        unchanged |= true;
+                    } else {
+                        return Err(e);
+                    }
+                },
+            }
         }
 
         if changed {
