@@ -4,25 +4,24 @@ use std::{
     fs,
     io::Read,
     path::{Path, PathBuf},
-    time::{Duration, SystemTime, UNIX_EPOCH},
+    time::{SystemTime, UNIX_EPOCH},
     sync::Arc,
 };
 
 use anyhow::{anyhow, Context, Result};
-
-// XXX: Requires fallible initialization, see https://github.com/rust-lang/rust/issues/109737
-use once_cell::unsync::OnceCell;
+use rusqlite::{
+    Connection,
+    OpenFlags,
+};
 
 use sequoia_openpgp as openpgp;
 use openpgp::{
     Fingerprint,
-    Cert,
-    cert::raw::RawCertParser,
+    cert::raw::{RawCert, RawCertParser},
     crypto::hash::Digest,
     KeyHandle,
     packet::UserID,
     parse::Parse,
-    serialize::SerializeInto,
     types::HashAlgorithm,
 };
 
@@ -37,7 +36,7 @@ use cert_store::store::UserIDQueryParams;
 use sequoia_wot as wot;
 
 use crate::{
-    common::Query,
+    common::{Common, Query},
     print_error_chain,
 };
 
@@ -58,7 +57,6 @@ pub struct KeyDB<'a> {
 struct Resource {
     kind: Kind,
     path: PathBuf,
-    create: bool,
 }
 
 #[derive(PartialEq, Eq, Clone, Copy)]
@@ -66,8 +64,8 @@ struct Resource {
 pub enum Kind {
     Keybox,
     KeyboxX509,
+    KeyboxDB,
     Keyring,
-    CertD,
 }
 
 impl Kind {
@@ -80,11 +78,6 @@ impl Kind {
         tracer!(TRACE, "Kind::guess");
         t!("Guessing kind of {:?}", path.as_ref());
 
-        if path.as_ref().is_dir() {
-            // XXX: Is there a more robust way to detect cert-ds?
-            return Ok(Some(Kind::CertD));
-        }
-
         let mut f = fs::File::open(path)?;
 
         // If the file is empty, GnuPG always returns keyring.
@@ -163,17 +156,15 @@ impl<'store> KeyDB<'store> {
         } else if url.starts_with("gnupg-kbx:") {
             kind = Some(Kind::Keybox);
             url = &url[10..];
-        } else if url.starts_with("pgp-cert-d:") {
-            kind = Some(Kind::CertD);
-            url = &url[11..];
+        } else if url.starts_with("gnupg-kbx-db:") {
+            kind = Some(Kind::KeyboxDB);
+            url = &url[13..];
         }
 
         // Expand tildes.
         let mut path = PathBuf::from(shellexpand::tilde(url).as_ref());
 
-        // If the path contains just a single component, it is
-        // relative to the home directory.
-        if path.components().count() == 1 {
+        if ! path.is_absolute() {
             path = home_dir.join(path);
         }
         t!("abolute path: {:?}", path);
@@ -245,7 +236,7 @@ impl<'store> KeyDB<'store> {
             None =>
                 Err(anyhow!("Unknown type of key resource {:?}", path)),
             Some(kind) => {
-                if ! create && ! path.exists() {
+                if ! create && ! path.exists() && kind != Kind::KeyboxDB {
                     return
                         Err(anyhow!("Key resource {:?} does not exist", path));
                 }
@@ -254,7 +245,6 @@ impl<'store> KeyDB<'store> {
                     Resource {
                         path,
                         kind,
-                        create,
                     }
                 );
                 Ok(())
@@ -266,7 +256,7 @@ impl<'store> KeyDB<'store> {
     ///
     /// Note: The returned certs have to be validated using a trust
     /// model!
-    pub fn lookup_candidates(&self, query: &Query)
+    pub fn lookup_candidates(&self, config: &dyn Common, query: &Query)
         -> Result<Vec<Arc<LazyCert<'store>>>>
     {
         tracer!(TRACE, "KeyDB::lookup_candidates");
@@ -278,7 +268,17 @@ impl<'store> KeyDB<'store> {
                 self.lookup_by_email(e),
             Query::UserIDFragment(f) =>
                 self.grep_userid(f),
-        }
+        }.map(|certs| {
+            for cert in &certs {
+                let _ = config.status().emit(
+                    crate::status::Status::KeyConsidered {
+                        fingerprint: cert.fingerprint(),
+                        not_selected: false,
+                        all_expired_or_revoked: false,
+                    });
+            }
+            certs
+        })
     }
 
     /// Adds a writable pgp-cert-d overlay to the resources, if not
@@ -299,113 +299,8 @@ impl<'store> KeyDB<'store> {
         self.overlay.as_ref().map_err(|_| anyhow::anyhow!("No overlay added"))
     }
 
-    // Initialize a certd.
-    fn initialize_certd<P>(&mut self, path: P, lazy: bool)
-        -> Result<Vec<(openpgp_cert_d::Tag, LazyCert<'store>)>>
-        where P: AsRef<Path>,
-    {
-        tracer!(TRACE, "KeyDB::initialize_certd");
-        let path = path.as_ref();
-        t!("loading cert-d {:?}", path);
-
-        let certd = openpgp_cert_d::CertD::with_base_dir(&path)
-            .map_err(|err| {
-                let err = anyhow::Error::from(err)
-                    .context(format!("While opening the certd {:?}", path));
-                print_error_chain(&err);
-                err
-            })?;
-
-        let items = certd.fingerprints();
-
-        let open = |fp: String| -> Option<(String, _, _)> {
-            let f = match certd.get_file(&fp) {
-                Ok(f) => f?,
-                Err(err) => {
-                    t!("Reading {}: {}", fp, err);
-                    return None;
-                }
-            };
-            match openpgp_cert_d::Tag::try_from(&f) {
-                Ok(tag) => Some((fp, tag, f)),
-                Err(err) => {
-                    t!("Getting tag for entry {}: {}", fp, err);
-                    None
-                }
-            }
-        };
-
-        let result = if lazy {
-            items.into_iter().filter_map(Result::ok).filter_map(|fp| {
-                // XXX: Once we have a cached tag, avoid the
-                // work if tags match.
-                t!("loading {} from overlay", fp);
-
-                let (fp, tag, file) = open(fp)?;
-
-                let mut parser = match RawCertParser::from_reader(file) {
-                    Ok(parser) => parser,
-                    Err(err) => {
-                        let err = anyhow::Error::from(err).context(format!(
-                            "While reading {:?} from the certd {:?}",
-                            fp, path));
-                        print_error_chain(&err);
-                        return None;
-                    }
-                };
-
-                match parser.next() {
-                    Some(Ok(cert)) => Some((tag, LazyCert::from(cert))),
-                    Some(Err(err)) => {
-                        let err = anyhow::Error::from(err).context(format!(
-                            "While parsing {:?} from the certd {:?}",
-                            fp, path));
-                        print_error_chain(&err);
-                        None
-                    }
-                    None => {
-                        let err = anyhow::anyhow!(format!(
-                            "While parsing {:?} from the certd {:?}: empty file",
-                            fp, path));
-                        print_error_chain(&err);
-                        None
-                    }
-                }
-            }).collect()
-        } else {
-            use rayon::prelude::*;
-
-            // For performance reasons, we read, parse, and
-            // canonicalize certs in parallel.
-            items.collect::<Vec<_>>().into_par_iter()
-                .filter_map(Result::ok)
-                .filter_map(|fp| {
-                    // XXX: Once we have a cached tag and
-                    // presumably a Sync index, avoid the work if
-                    // tags match.
-                    t!("loading {} from overlay", fp);
-
-                    let (fp, tag, file) = open(fp)?;
-
-                    match Cert::from_reader(file) {
-                        Ok(cert) => Some((tag, LazyCert::from(cert))),
-                        Err(err) => {
-                            let err = anyhow::Error::from(err).context(format!(
-                                "While parsing {:?} from the certd {:?}",
-                                fp, path));
-                            print_error_chain(&err);
-                            None
-                        }
-                    }
-                })
-                .collect()
-        };
-
-        Ok(result)
-    }
-
     // Initialize a keyring.
-    fn initialize_keyring<P>(&mut self, file: fs::File, path: P, lazy: bool)
+    fn initialize_keyring<P>(&mut self, file: fs::File, path: P)
         -> Result<Vec<LazyCert<'store>>>
         where P: AsRef<Path>,
     {
@@ -413,7 +308,7 @@ impl<'store> KeyDB<'store> {
         let path = path.as_ref();
         t!("loading keyring {:?}", path);
 
-        let results = if lazy {
+        let results = {
             let iter = match RawCertParser::from_reader(file) {
                 Ok(iter) => iter,
                 Err(err) => {
@@ -435,35 +330,13 @@ impl<'store> KeyDB<'store> {
                     }
                 }
             }).collect()
-        } else {
-            let iter = match sequoia_openpgp_mt::keyring::parse(file) {
-                Ok(iter) => iter,
-                Err(err) => {
-                    let err = anyhow::Error::from(err).context(
-                        format!("Loading keyring {:?}", path));
-                    print_error_chain(&err);
-                    return Err(err);
-                }
-            };
-
-            iter.into_iter().filter_map(|cert| {
-                match cert {
-                    Ok(cert) => Some(LazyCert::from(cert)),
-                    Err(err) => {
-                        let err = anyhow::Error::from(err).context(format!(
-                            "While parsing cert from keyring {:?}", path));
-                        print_error_chain(&err);
-                        None
-                    }
-                }
-            }).collect()
         };
 
         Ok(results)
     }
 
     // Initialize a keybox.
-    fn initialize_keybox<P>(&mut self, file: fs::File, path: P, _lazy: bool)
+    fn initialize_keybox<P>(&mut self, file: fs::File, path: P)
         -> Result<Vec<LazyCert<'store>>>
         where P: AsRef<Path>,
     {
@@ -512,6 +385,35 @@ impl<'store> KeyDB<'store> {
         Ok(results)
     }
 
+    /// Initialize a keybox database.
+    fn initialize_keybox_db<P>(&mut self, path: P)
+        -> Result<Vec<LazyCert<'store>>>
+        where P: AsRef<Path>,
+    {
+        tracer!(TRACE, "KeyDB::initialize_keybox_db");
+        let path = path.as_ref();
+        t!("loading keybox database at {}", path.display());
+
+        let conn = Connection::open_with_flags(
+            &path, OpenFlags::SQLITE_OPEN_READ_ONLY)?;
+
+        let mut stmt = conn.prepare("SELECT keyblob \
+                                     FROM pubkey \
+                                     WHERE type = 1")?;
+
+        let certs = stmt.query_map([], |row| Ok(row.get::<_, Vec<u8>>(0)?))?
+            .filter_map(|bytes| {
+                let bytes = std::io::Cursor::new(bytes.ok()?);
+                let cert = RawCert::from_reader(bytes).ok()?;
+                t!("loaded {}", cert.fingerprint());
+                Some(cert.into())
+            })
+            .collect();
+        drop(stmt);
+
+        Ok(certs)
+    }
+
     /// Initializes the store, if not already done.
     #[allow(dead_code)]
     pub fn initialize(&mut self, lazy: bool) -> Result<()> {
@@ -536,13 +438,17 @@ impl<'store> KeyDB<'store> {
         t!("initializing");
 
         for resource in &self.resources.clone() {
-            if resource.create && ! resource.path.exists() {
-                t!("skipping non-existing resource {:?}", resource.path);
+            if ! resource.path.exists() {
+                t!("{}: skipping non-existing resource", resource.path.display());
                 continue;
             }
 
-            let f = fs::File::open(&resource.path)?;
-            let modified = f.metadata()?.modified()?;
+            let f = fs::File::open(&resource.path);
+            let modified = match &f {
+                Ok(f) => Some(f.metadata()?.modified()?),
+                Err(_) => None,
+            };
+            t!("{}: last modified {:?}", resource.path.display(), modified);
 
             // If there is a writable openpgp-cert-d overlay on top of
             // the stack.  We import all certs from our resources
@@ -556,23 +462,23 @@ impl<'store> KeyDB<'store> {
 
             if self.overlay.as_ref().ok()
                 .and_then(|overlay| overlay.get_cached_mtime(&resource).ok())
-                .map(|cached| unix_time(modified) == unix_time(cached))
+                .map(|cached| modified.map(unix_time) == Some(unix_time(cached)))
                 .unwrap_or(false)
             {
                 // The overlay already contains all data from
                 // this resource.
-                t!("skipping up-to-date resource {:?}", resource.path);
+                t!("{}: skipping up-to-date resource", resource.path.display());
                 continue;
             }
 
             let certs = match resource.kind {
                 Kind::Keyring => {
-                    self.initialize_keyring(f, &resource.path, lazy)
+                    self.initialize_keyring(f?, &resource.path)
                         .with_context(|| format!(
                             "Reading the keyring {:?}", resource.path))
                 },
                 Kind::Keybox => {
-                    self.initialize_keybox(f, &resource.path, lazy)
+                    self.initialize_keybox(f?, &resource.path)
                         .with_context(|| format!(
                             "Reading the keybox {:?}", resource.path))
                 },
@@ -581,17 +487,12 @@ impl<'store> KeyDB<'store> {
                        resource.path);
                     Ok(Vec::new())
                 },
-                Kind::CertD => {
-                    self.initialize_certd(&resource.path, lazy)
+                Kind::KeyboxDB => {
+                    self.initialize_keybox_db(&resource.path)
                         .with_context(|| format!(
-                            "Reading the certd {:?}", resource.path))
-                        .map(|certs| {
-                            certs
-                                .into_iter()
-                                .map(|(_tag, cert)| cert)
-                                .collect()
-                        })
-                }
+                            "{}: reading the keybox database",
+                            resource.path.display()))
+                },
             };
 
             match certs {
@@ -611,7 +512,7 @@ impl<'store> KeyDB<'store> {
                 Err(err) => print_error_chain(&err),
             }
 
-            if let Ok(overlay) = &self.overlay {
+            if let (Ok(overlay), Some(modified)) = (&self.overlay, modified) {
                 overlay.set_cached_mtime(&resource, modified)?;
             }
         }
@@ -629,8 +530,6 @@ impl<'store> KeyDB<'store> {
 
 pub struct Overlay<'store> {
     pub(crate) cert_store: CertStore<'store>,
-    #[allow(dead_code)]
-    trust_root: OnceCell<Cert>,
 }
 
 impl<'store> Overlay<'store> {
@@ -667,7 +566,6 @@ impl<'store> Overlay<'store> {
 
         Ok(Overlay {
             cert_store,
-            trust_root: Default::default(),
         })
     }
 
@@ -679,77 +577,11 @@ impl<'store> Overlay<'store> {
     }
 
     /// Lazily reads (or creates) the trust root.
-    pub fn trust_root(&self) -> Result<&Cert> {
-        self.trust_root.get_or_try_init(|| self.load_trust_root())
-    }
-
-    /// Eagerly loads the trust root, or generates one if none existed
-    /// before.
-    ///
-    /// This is done during the insertion, while we hold the exclusive
-    /// lock, so this is race free.
-    fn load_trust_root(&self) -> Result<Cert> {
-        use openpgp_cert_d::MergeResult;
-
-        let certd = self.certd();
-
-        // Acquire an exclusive lock on the certd by inserting the
-        // trust root.  We may yet discover that one already exists,
-        // in which case we won't generate a new one, but load the
-        // existing one.
-        let mut trust_root =
-            Err(anyhow::anyhow!("merge callback not invoked"));
-        certd.insert_special(
-            openpgp_cert_d::TRUST_ROOT,
-            (), false, |_new, old| {
-                if let Some(old) = old {
-                    trust_root = Cert::from_bytes(&old);
-                    Ok(MergeResult::Keep)
-                } else {
-                    let tr = Self::generate_trust_root()?;
-                    let d = tr.as_tsk().to_vec()?;
-                    trust_root = Ok(tr);
-                    Ok(MergeResult::Data(d))
-                }
-            })?;
-        let trust_root = trust_root?;
-
-        // Also insert the public bits into the certd for the WoT
-        // algorithm to find.
-        certd.insert(
-            &trust_root.fingerprint().to_string(), &trust_root, false,
-            |new, _old| {
-                Ok(openpgp_cert_d::MergeResult::Data(new.to_vec()?))
-            })?;
-
-        Ok(trust_root)
-    }
-
-    fn generate_trust_root() -> Result<Cert> {
-        use openpgp::{
-            cert::CertBuilder,
-            packet::signature::SignatureBuilder,
-            types::SignatureType,
-        };
-
-        // XXX: It would be nice if the direct key signature would
-        // also be non-exportable, but Sequoia doesn't have a way to
-        // do that yet with the CertBuilder.
-        let (root, _) =
-            CertBuilder::new()
-            // Set it in the past so that it is possible to use the CA
-            // when the reference time is in the past.  Feb 2002.
-            .set_creation_time(
-                SystemTime::UNIX_EPOCH + Duration::new(1014235320, 0))
-            // CAs should *not* expire.
-            .set_validity_period(None)
-            .add_userid_with(
-                "Local Trust Root",
-                SignatureBuilder::new(SignatureType::PositiveCertification)
-                    .set_exportable_certification(false)?)?
-            .generate()?;
-
-        Ok(root)
+    pub fn trust_root(&self) -> Result<Arc<LazyCert<'store>>> {
+        self.cert_store
+            .certd().expect("created using CertStore::open")
+            .trust_root()
+            .map(move |(cert, _created)| cert)
     }
 
     pub fn path(&self) -> &Path {
@@ -797,14 +629,6 @@ macro_rules! forward {
         }
     }}
 }
-macro_rules! forward_mut {
-    ( $method:ident, $self:expr $(, $args:ident)* ) => {{
-        match $self.overlay.as_mut() {
-            Ok(be) => be.cert_store.$method($($args),*),
-            Err(be) => be.$method($($args),*),
-        }
-    }}
-}
 
 impl<'a> cert_store::store::Store<'a> for KeyDB<'a> {
     fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
@@ -856,25 +680,25 @@ impl<'a> cert_store::store::Store<'a> for KeyDB<'a> {
         forward!(certs, self)
     }
 
-    fn prefetch_all(&mut self) {
-        forward_mut!(prefetch_all, self)
+    fn prefetch_all(&self) {
+        forward!(prefetch_all, self)
     }
 
-    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
-        forward_mut!(prefetch_some, self, certs)
+    fn prefetch_some(&self, certs: &[KeyHandle]) {
+        forward!(prefetch_some, self, certs)
     }
 }
 
 impl<'a> cert_store::store::StoreUpdate<'a> for KeyDB<'a> {
-    fn update(&mut self, cert: Arc<LazyCert<'a>>) -> Result<()> {
-        forward_mut!(update, self, cert)
+    fn update(&self, cert: Arc<LazyCert<'a>>) -> Result<()> {
+        forward!(update, self, cert)
     }
 
-    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
-                      merge_strategy: &mut dyn MergeCerts<'a>)
+    fn update_by(&self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &dyn MergeCerts<'a>)
         -> Result<Arc<LazyCert<'a>>>
     {
-        forward_mut!(update_by, self, cert, merge_strategy)
+        forward!(update_by, self, cert, merge_strategy)
     }
 }
 
