@@ -1,5 +1,10 @@
 use std::{
-    collections::BTreeMap,
+    cell::OnceCell,
+    collections::{
+        BTreeMap,
+        BTreeSet,
+        HashSet,
+    },
     io::{self, Write},
     sync::Arc,
 };
@@ -8,8 +13,18 @@ use anyhow::Result;
 
 use sequoia_openpgp as openpgp;
 use openpgp::{
-    cert::amalgamation::{ValidateAmalgamation, ValidAmalgamation},
-    packet::{UserID, signature::subpacket::SubpacketTag},
+    Cert,
+    Fingerprint,
+    cert::{
+        amalgamation::{ValidateAmalgamation, ValidAmalgamation},
+        raw::RawCertParser,
+    },
+    packet::{
+        Key,
+        Signature,
+        key,
+    },
+    parse::Parse,
     types::*,
 };
 use sequoia_ipc as ipc;
@@ -114,31 +129,31 @@ impl ListOptions {
             "show key usage information during key listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-policy-urls",
             |o, s, _| Ok({ o.policy_urls = s; }),
             "show policy URLs during signature listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-notations",
             |o, s, _| Ok({ o.ietf_notations = s; o.user_notations = s; }),
             "show all notations during signature listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-std-notations",
             |o, s, _| Ok({ o.ietf_notations = s; }),
             "show IETF standard notations during signature listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-standard-notations",
             |o, s, _| Ok({ o.ietf_notations = s; }),
             "",
         },
 
-        opt_todo! {
+        opt! {
             "show-user-notations",
             |o, s, _| Ok({ o.user_notations = s; }),
             "show user-supplied notations during signature listings",
@@ -156,7 +171,7 @@ impl ListOptions {
             "show user ID validity during key listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-unusable-uids",
             |o, s, _| Ok({ o.unusable_uids = s; }),
             "show revoked and expired user IDs in key listings",
@@ -246,8 +261,6 @@ pub fn cmd_list_keys(config: &crate::Config, args: &[String], list_secret: bool)
                     overlay.cert_store.certd().map(|c| c.certd())
                 {
                     use openpgp::{
-                        Cert,
-                        parse::Parse,
                         serialize::SerializeInto,
                     };
                     use sequoia_cert_store::store::openpgp_cert_d::MergeResult;
@@ -256,7 +269,7 @@ pub fn cmd_list_keys(config: &crate::Config, args: &[String], list_secret: bool)
                     {
                         certd.insert(
                             &trust_root.fingerprint().to_string(),
-                            trust_root, false,
+                            trust_root.to_cert()?, false,
                             |new, old| {
                                 let d = if let Some(old) = old
                                     .and_then(|b| Cert::from_bytes(b).ok())
@@ -285,8 +298,8 @@ pub fn cmd_list_keys(config: &crate::Config, args: &[String], list_secret: bool)
                         .unwrap_or(true)))
     } else {
         let mut certs = BTreeMap::new();
-        for query in args.iter().map(|a| Query::from(&a[..])) {
-            let r = match query {
+        for query in args.iter().map(|a| a.parse()) {
+            let r = match query? {
                 Query::Key(h) | Query::ExactKey(h) =>
                     config.keydb().lookup_by_cert_or_subkey(&h),
                 Query::Email(e) =>
@@ -319,14 +332,70 @@ pub fn cmd_list_keys(config: &crate::Config, args: &[String], list_secret: bool)
     };
 
     list_keys(config, certs, list_secret,
-              args.is_empty(), // Only print header if no query is given.
+              args.is_empty(), // Are we listing all certs?
               sink)
 }
 
+/// Dispatches the --show-keys command.
+///
+/// Note: contrary to GnuPG we don't use the import code.
+pub fn cmd_show_keys(config: &crate::Config, args: &[String])
+                     -> Result<()>
+{
+    let sink = io::stdout(); // XXX
+
+    let mut certs = Vec::new();
+
+    // If the list of arguments is empty, stdin is implied.
+    let stdin: Vec<String> = vec!["-".to_string()];
+    let filenames = if args.is_empty() {
+        &stdin
+    } else {
+        args
+    };
+
+    // Parse every cert from every file.
+    for filename in filenames {
+        // XXX: Would be nice to mmap the file.
+        let parser = match crate::utils::open(config, &filename)
+            .and_then(|f| RawCertParser::from_reader(f))
+        {
+            Ok(c) => c,
+            Err(e) => {
+                config.warn(format_args!("can't open '{}': {}", filename, e));
+                continue;
+            },
+        };
+
+        for cert in parser {
+            match cert {
+                Ok(c) => certs.push(Arc::new(c.into())),
+                Err(e) => {
+                    config.warn(format_args!("skipping key in '{}': {}",
+                                             filename, e));
+                },
+            }
+        }
+    }
+
+    if certs.is_empty() {
+        return Err(anyhow::anyhow!(
+            "error reading key: No public key"));
+    }
+
+    let rt = tokio::runtime::Runtime::new()?;
+    rt.block_on(async_list_keys(config, certs.into_iter(),
+                                true, false,
+                                false,
+                                true,
+                                false,
+                                sink))
+}
+
 pub fn list_keys<'a, 'store: 'a, S>(config: &'a crate::Config<'store>,
                                     certs: impl Iterator<Item = Arc<LazyCert<'store>>>,
                                     list_secret: bool,
-                                    emit_header: bool,
+                                    list_all: bool,
                                     sink: S)
     -> Result<()>
 where
@@ -336,7 +405,7 @@ where
     rt.block_on(async_list_keys(config, certs,
                                 list_secret, list_secret,
                                 config.list_options.uid_validity,
-                                emit_header,
+                                list_all, list_all,
                                 sink))
 }
 
@@ -347,32 +416,49 @@ pub async fn async_list_keys<'a, 'store: 'a, S>(
     // Tunes behavior for gpg --list-secret-keys.
     list_secret_keys_mode: bool,
     list_uid_validity: bool,
-    emit_header: bool,
+    list_all: bool,
+    list_header: bool,
     mut sink: S)
     -> Result<()>
 where
     S: Write,
 {
-    let vtm = config.trust_model_impl.with_policy(config, Some(config.now()))?;
+    let vtm = config.trust_model_impl.with_policy_and_precompute(
+        config, Some(config.now()), list_all && ! list_secret_keys_mode)?;
     let p = vtm.policy();
 
-    let mut agent =
-        if list_secret || (config.with_secret && config.with_colons) {
-            config.connect_agent().await.ok()
-        } else {
-            None
-        };
+    let mut secrets = Default::default();
+    if list_secret || (config.with_secret && config.with_colons) {
+        if let Ok(mut agent) = config.connect_agent().await {
+            secrets = agent.list_keys().await
+                .map(|keys| {
+                    keys.into_iter()
+                        .map(|k| k.keygrip().clone())
+                        .collect::<HashSet<_>>()
+                })
+                .unwrap_or(Default::default())
+        }
+    }
 
     // We emit the location header for humans only if we actually list
     // at least one key.
     let mut emitted_header = false;
 
+    // For --check-sigs, we compute some stats.
+    let mut sig_stats = SigStats::default();
+
     for cert in certs {
-        let mut has_secret = if let Some(a) = agent.as_mut() {
-            crate::agent::has_keys(a, &cert).await?
-        } else {
-            Default::default()
-        };
+        let mut has_secret: BTreeSet<Fingerprint> = cert
+            .keys()
+            .filter_map(|k| {
+                let keygrip = Keygrip::of(k.mpis()).ok()?;
+                if secrets.contains(&keygrip) {
+                    Some(k.fingerprint())
+                } else {
+                    None
+                }
+            })
+            .collect();
 
         // When we are importing secret keys, we may have the secret
         // while the agent does not yet have it.  Nevertheless, we
@@ -392,9 +478,10 @@ where
             continue;
         };
 
-        // For humans, we print the location of the store if we list
+        // For humans, we print the location of the store if the user
+        // requested all keys (i.e. no pattern was given) and we list
         // at least one key.
-        if emit_header && ! emitted_header && ! config.with_colons {
+        if list_header && ! emitted_header && ! config.with_colons {
             emitted_header = true;
 
             let path =
@@ -407,12 +494,9 @@ where
         let acert = AuthenticatedCert::new(vtm.as_ref(), &cert)?;
         let vcert = cert.with_policy(p, config.now()).ok();
         let cert_fp = cert.fingerprint();
-        let cert_kh = cert.key_handle();
         let have_secret = has_secret.contains(&cert_fp);
         let ownertrust = config.trustdb.get_ownertrust(&cert_fp)
             .unwrap_or_else(|| OwnerTrustLevel::Undefined.into());
-        let best_effort_primary_userid: UserID =
-            best_effort_primary_uid(config.policy(), &cert).into();
 
         Record::Key {
             key: cert.primary_key().key(),
@@ -484,43 +568,25 @@ where
 
         if config.list_options.list_sigs {
             for s in cert.primary_key().signatures() {
-                let is_self_sig =
-                    s.get_issuers().iter().any(|i| i.aliases(&cert_kh));
-
                 let (issuer_uid, validity) =
-                    if config.list_options.fast_list {
-                        (Some("".into()), None)
-                    } else if is_self_sig {
-                        (Some(best_effort_primary_userid.clone()), None)
-                    } else {
-                        if let Some(signer) =
-                            s.get_issuers().into_iter().find_map(
-                                |k| config.keydb().lookup_by_cert_or_subkey(&k).ok())
-                            .and_then(|certs| certs.into_iter().next())
-                            .and_then(|cert| cert.to_cert().ok().cloned())
-                        {
-                            (Some(best_effort_primary_uid(config.policy(), &signer).into()),
-                             None)
+                    compute_sig_issuer_uid_and_validity(
+                        config, &mut sig_stats,
+                        cert, s,
+                        |k| if s.typ() == SignatureType::KeyRevocation {
+                            s.clone().verify_primary_key_revocation(
+                                k, cert.primary_key().key())
                         } else {
-                            (None, Some(SignatureValidity::MissingKey))
-                        }
-                    };
+                            s.clone().verify_direct_key(
+                                k, cert.primary_key().key())
+                        });
+                if validity.suppress(config) {
+                    continue; // Skip this signature.
+                }
 
                 Record::Signature {
-                    issuer: s.issuers().cloned().next().or_else(
-                        || s.issuer_fingerprints().cloned().next().map(Into::into)),
-                    issuer_fp: s.issuer_fingerprints().cloned().next(),
+                    sig: s,
                     issuer_uid,
                     validity,
-                    pk_algo: s.pk_algo(),
-                    hash_algo: s.hash_algo(),
-                    creation_time: s.signature_creation_time()
-                        .expect("valid signatures have a creation time"),
-                    typ: s.typ(),
-                    exportable: s.exportable().is_ok(),
-                    trust: s.trust_signature(),
-                    has_notations: s.subpackets(SubpacketTag::NotationData)
-                        .next().is_some(),
                 }.emit(config, &mut sink)?;
             }
         }
@@ -537,11 +603,27 @@ where
         for (validity, uid) in userids.into_iter() {
             let vuid = uid.clone().with_policy(p, config.now()).ok();
 
-            if ! vuid.is_some() {
-                continue;
+            // Unless explicitly requested, we don't list unusable
+            // user ids.
+            if ! config.list_options.unusable_uids {
+                // We never list user ids without any self signatures.
+                if uid.self_signatures().count() == 0 {
+                    continue;
+                }
+
+                // In human readable mode, we don't list expired user
+                // ids, or those that are revoked (unless the whole
+                // cert is revoked).
+                if ! config.with_colons
+                    && ((validity.revoked && ! acert.cert_validity().revoked)
+                        || vuid.is_none())
+                {
+                    continue;
+                }
             }
 
             Record::UserID {
+                amalgamation: uid.clone(),
                 validity: (list_uid_validity
                            // For some reason, in the machine readable
                            // output, GnuPG disregards
@@ -553,57 +635,30 @@ where
                                    || validity.expired
                                    || ownertrust.disabled())))
                     .then_some(validity),
-                creation_date: vuid.as_ref()
-                    .and_then(|v| v.binding_signature().signature_creation_time())
-                    .unwrap_or_else(|| {
-                        uid.self_signatures().next()
-                            .and_then(|s| s.signature_creation_time())
-                            .unwrap_or(std::time::UNIX_EPOCH)
-                    }),
-                expiration_date:  vuid.as_ref()
-                    .and_then(|v| v.binding_signature().signature_expiration_time()),
-                userid: uid.userid().clone(),
             }.emit(config, &mut sink)?;
 
             if config.list_options.list_sigs {
                 for s in uid.signatures() {
-                    let is_self_sig =
-                        s.get_issuers().iter().any(|i| i.aliases(&cert_kh));
-
                     let (issuer_uid, validity) =
-                        if config.list_options.fast_list {
-                            (Some("".into()), None)
-                        } else if is_self_sig {
-                            (Some(best_effort_primary_userid.clone()), None)
-                        } else {
-                            if let Some(signer) =
-                                s.get_issuers().into_iter().find_map(
-                                    |k| config.keydb().lookup_by_cert_or_subkey(&k).ok())
-                                .and_then(|certs| certs.into_iter().next())
-                                .and_then(|cert| cert.to_cert().ok().cloned())
-                            {
-                                (Some(best_effort_primary_uid(config.policy(), &signer).into()),
-                                 None)
+                        compute_sig_issuer_uid_and_validity(
+                            config, &mut sig_stats,
+                            cert, s,
+                            |k| if s.typ() == SignatureType::CertificationRevocation {
+                                s.clone().verify_userid_revocation(
+                                    k, cert.primary_key().key(), uid.userid())
                             } else {
-                                (None, Some(SignatureValidity::MissingKey))
-                            }
-                        };
+                                s.clone().verify_userid_binding(
+                                    k, cert.primary_key().key(), uid.userid())
+                            });
+
+                    if validity.suppress(config) {
+                        continue; // Skip this signature.
+                    }
 
                     Record::Signature {
-                        issuer: s.issuers().cloned().next().or_else(
-                            || s.issuer_fingerprints().cloned().next().map(Into::into)),
-                        issuer_fp: s.issuer_fingerprints().cloned().next(),
+                        sig: s,
                         issuer_uid,
                         validity,
-                        pk_algo: s.pk_algo(),
-                        hash_algo: s.hash_algo(),
-                        creation_time: s.signature_creation_time()
-                            .expect("valid signatures have a creation time"),
-                        typ: s.typ(),
-                        exportable: s.exportable().is_ok(),
-                        trust: s.trust_signature(),
-                        has_notations: s.subpackets(SubpacketTag::NotationData)
-                            .next().is_some(),
                     }.emit(config, &mut sink)?;
                 }
             }
@@ -656,43 +711,26 @@ where
 
             if config.list_options.list_sigs {
                 for s in subkey.signatures() {
-                    let is_self_sig =
-                        s.get_issuers().iter().any(|i| i.aliases(&cert_kh));
-
                     let (issuer_uid, validity) =
-                        if config.list_options.fast_list {
-                            (Some("".into()), None)
-                        } else if is_self_sig {
-                            (Some(best_effort_primary_userid.clone()), None)
-                        } else {
-                            if let Some(signer) =
-                                s.get_issuers().into_iter().find_map(
-                                    |k| config.keydb().lookup_by_cert_or_subkey(&k).ok())
-                                .and_then(|certs| certs.into_iter().next())
-                                .and_then(|cert| cert.to_cert().ok().cloned())
-                            {
-                                (Some(best_effort_primary_uid(config.policy(), &signer).into()),
-                                 None)
+                        compute_sig_issuer_uid_and_validity(
+                            config, &mut sig_stats,
+                            cert, s,
+                            |k| if s.typ() == SignatureType::SubkeyRevocation {
+                                s.clone().verify_subkey_revocation(
+                                    k, cert.primary_key().key(), subkey.key())
                             } else {
-                                (None, Some(SignatureValidity::MissingKey))
-                            }
-                        };
+                                s.clone().verify_subkey_binding(
+                                    k, cert.primary_key().key(), subkey.key())
+                            });
+
+                    if validity.suppress(config) {
+                        continue; // Skip this signature.
+                    }
 
                     Record::Signature {
-                        issuer: s.issuers().cloned().next().or_else(
-                            || s.issuer_fingerprints().cloned().next().map(Into::into)),
-                        issuer_fp: s.issuer_fingerprints().cloned().next(),
+                        sig: s,
                         issuer_uid,
                         validity,
-                        pk_algo: s.pk_algo(),
-                        hash_algo: s.hash_algo(),
-                        creation_time: s.signature_creation_time()
-                            .expect("valid signatures have a creation time"),
-                        typ: s.typ(),
-                        exportable: s.exportable().is_ok(),
-                        trust: s.trust_signature(),
-                        has_notations: s.subpackets(SubpacketTag::NotationData)
-                            .next().is_some(),
                     }.emit(config, &mut sink)?;
                 }
             }
@@ -704,5 +742,153 @@ where
         }
     }
 
+    sig_stats.emit(config);
     Ok(())
 }
+
+impl SignatureValidity {
+    /// Returns whether this signature should be skipped in the
+    /// signature listing.
+    pub fn suppress(&self, config: &crate::Config) -> bool {
+        config.check_sigs && ! config.with_colons
+            && (self == &SignatureValidity::NotChecked
+                || self == &SignatureValidity::MissingKey)
+    }
+}
+
+/// Computes IssuerUserID and SignatureValidity for the given
+/// signature.
+///
+/// This computes various checks on the signature, depending on the
+/// options given.  It also encapsulates some of the complexity with
+/// matching GnuPG's output.
+///
+/// Notably, GnuPG has two different sets of functions, one for the
+/// human-readable output (list_signature_print), and one for the
+/// machine-readable output (list_keyblock_colon).  What is computed
+/// and what is shown deviates slightly.  On the other hand, we have a
+/// unified output function, and capture the difference here in this
+/// function.
+fn compute_sig_issuer_uid_and_validity<C>(config: &crate::Config,
+                                          sig_stats: &mut SigStats,
+                                          cert: &Cert,
+                                          s: &Signature,
+                                          mut check_sig: C)
+                                          -> (IssuerUserID, SignatureValidity)
+where
+    C: FnMut(&Key<key::PublicParts, key::UnspecifiedRole>) -> Result<()>,
+{
+    let cert_kh = cert.key_handle();
+    let is_self_sig =
+        s.get_issuers().iter().any(|i| i.aliases(&cert_kh));
+
+    // Lazily look up the signer cert.
+    //
+    // XXX: Currently, we only consider the first candidate.  It'd be
+    // better to consider all, then set the signers_uid to the first
+    // cert that successfully verifies the signature.
+    let signer = OnceCell::<Option<Cert>>::new();
+    let lookup_signer = |config: &crate::Config, sig: &Signature| {
+        signer.get_or_init(|| {
+            sig.get_issuers().into_iter()
+                .find_map(
+                    |k| config.lookup_by_cert_or_subkey(&k).ok())
+                .and_then(|certs| certs.into_iter().next())
+                .and_then(|cert| cert.to_cert().ok().cloned())
+        })
+    };
+
+    // Compute the issuer certificates (best effort) primary user ID
+    // for display.
+    let issuer_uid = if config.check_sigs && s.signature_creation_time()
+        .map(|sct| sct < cert.primary_key().creation_time())
+        .unwrap_or(false)
+    {
+        IssuerUserID::TimeConflict
+    } else if config.list_options.fast_list {
+        IssuerUserID::Empty
+    } else if is_self_sig && config.list_options.fast_list {
+        IssuerUserID::Empty
+    } else if let Some(signer) = lookup_signer(config, s) {
+        best_effort_primary_uid(config.policy(), &signer).into()
+    } else {
+        IssuerUserID::NotFound
+    };
+
+    // Now compute the signature validity.
+    let sig_validity = if config.check_sigs && s.signature_creation_time()
+        .map(|sct| sct < cert.primary_key().creation_time())
+        .unwrap_or(false)
+    {
+        config.warn(format_args!(
+            "public key {:X} is {} seconds newer than the signature",
+            cert.keyid(),
+            (cert.primary_key().creation_time().duration_since(
+                s.signature_creation_time().unwrap()).unwrap().as_secs())));
+        sig_stats.errors += 1;
+        SignatureValidity::OtherError
+    } else if ! config.check_sigs && config.list_options.fast_list {
+        sig_stats.missing_key += 1;
+        SignatureValidity::NotChecked
+    } else if let Some(signer) = lookup_signer(config, s) {
+        if config.check_sigs {
+            if signer.keys().key_handles(s.get_issuers().iter())
+                .any(|k| check_sig(k.key()).is_ok())
+            {
+                // XXX: check certification key flag above?
+                sig_stats.good += 1;
+                SignatureValidity::Good
+            } else {
+                sig_stats.bad += 1;
+                SignatureValidity::Bad
+            }
+        } else {
+            SignatureValidity::NotChecked
+        }
+    } else {
+        sig_stats.missing_key += 1;
+        SignatureValidity::MissingKey
+    };
+
+    (issuer_uid, sig_validity)
+}
+
+/// Signature statistics for --check-sigs.
+#[derive(Default)]
+struct SigStats {
+    good: usize,
+    bad: usize,
+    missing_key: usize,
+    errors: usize,
+}
+
+impl SigStats {
+    /// Emits statistics, if appropriate.
+    fn emit(&self, config: &crate::Config) {
+        if config.check_sigs && ! config.with_colons {
+            if self.good > 0 {
+                config.warn(format_args!(
+                    "{} good signatures",
+                    self.good));
+            }
+
+            if self.bad > 0 {
+                config.warn(format_args!(
+                    "{} bad signatures",
+                    self.bad));
+            }
+
+            if self.missing_key > 0 {
+                config.warn(format_args!(
+                    "{} signatures not checked due to missing keys",
+                    self.missing_key));
+            }
+
+            if self.errors > 0 {
+                config.warn(format_args!(
+                    "{} signatures not checked due to errors",
+                    self.errors));
+            }
+        }
+    }
+}
