@@ -19,7 +19,10 @@ use openpgp::{
     KeyID,
     crypto::{S2K, SessionKey},
     fmt::hex,
-    packet::UserID,
+    packet::{
+        PKESK,
+        UserID,
+    },
     types::*,
 };
 
@@ -43,17 +46,19 @@ impl<S: io::Write + Send + Sync + 'static> From<S> for Fd {
 }
 
 impl Fd {
-    // Sends all output to /dev/null.
+    /// Sends all output to /dev/null.
     pub fn sink() -> Self {
         Fd(None)
     }
 
-    // Whether gpg was called with --status-fd.  If not, everything is
-    // sent to /dev/null.
+    /// Whether gpg was called with --status-fd.
+    ///
+    /// If not, everything is sent to /dev/null.
     pub fn enabled(&self) -> bool {
         self.0.is_some()
     }
 
+    /// Emits a status message.
     #[allow(dead_code)]
     pub fn emit(&self, status: Status<'_>) -> Result<()> {
         crate::with_invocation_log(|sink| status.emit(sink));
@@ -64,10 +69,12 @@ impl Fd {
         }
     }
 
-    // If --status-fd was passed, then emits something on the
-    // specified file descriptor, otherwise prints the message plus a
-    // newline on stdout.
-    pub fn emit_or(&self, status: Status<'_>, msg: &str) -> Result<()> {
+    /// Emits a status message or an interactive prompt.
+    ///
+    /// If --status-fd was passed, then emits something on the
+    /// specified file descriptor, otherwise prints the message plus a
+    /// newline on stdout.
+    pub fn emit_or_prompt(&self, status: Status<'_>, msg: &str) -> Result<()> {
         crate::with_invocation_log(|sink| status.emit(sink));
         if self.enabled() {
             self.emit(status)
@@ -128,11 +135,15 @@ pub enum Status<'a> {
         creation_time: SystemTime,
     },
 
+    AssertPubkeyAlgo {
+        fp: Fingerprint,
+        accepted: bool,
+        policy: String,
+    },
+
     // Encryption-related.
     EncTo {
-        keyid: KeyID,
-        pk_algo: Option<PublicKeyAlgorithm>,
-        pk_len: Option<usize>,
+        pkesk: &'a PKESK,
     },
     BeginDecryption,
     EndDecryption,
@@ -288,15 +299,27 @@ pub enum Status<'a> {
 
     NoData(NoDataReason),
 
-    // What to prompt for, e.g., `ask_revocation_reason.okay`.
+    /// What to prompt for when asking a yes/no question, e.g.,
+    /// `ask_revocation_reason.okay`.
     GetBool(String),
 
-    // What to prompt for, e.g. `ask_revocation_reason.text`.
+    /// What to prompt for when asking a question,
+    /// e.g. `ask_revocation_reason.text`.
     GetLine(String),
 
-    // Sent when we got a prompt.
+    /// What to prompt for when asking for a password,
+    /// e.g. `passphrase.enter`.
+    GetHidden(String),
+
+    /// Sent when we got an answer.
     GotIt,
 
+    /// Indicates the maximum length of the requested input.
+    ///
+    /// Note: We don't have that limitation, and I'm not sure GnuPG
+    /// has.
+    InquireMaxLen(usize),
+
     Failure {
         location: &'a str,
         error: crate::error_codes::Error,
@@ -452,15 +475,28 @@ impl Status<'_> {
                          t.format("%s"))?;
             },
 
-            EncTo {
-                keyid,
-                pk_algo,
-                pk_len,
+            AssertPubkeyAlgo {
+                fp,
+                accepted,
+                policy,
             } => {
+                writeln!(w, "ASSERT_PUBKEY_ALGO {:X} {} {}",
+                         fp,
+                         if *accepted { 1 } else { 0 },
+                         policy)?;
+            },
+
+            EncTo {
+                pkesk,
+            } => {
+                // According to doc/DETAILS, GnuPG always reports the
+                // length as 0.
+                let pk_len = 0;
+
                 writeln!(w, "ENC_TO {:X} {} {}",
-                         keyid,
-                         pk_algo.map(|a| u8::from(a)).unwrap_or(0),
-                         pk_len.unwrap_or(0))?;
+                         pkesk.recipient(),
+                         u8::from(pkesk.pk_algo()),
+                         pk_len)?;
             },
 
             BeginDecryption => writeln!(w, "BEGIN_DECRYPTION")?,
@@ -814,13 +850,17 @@ impl Status<'_> {
 
             GetLine(prompt) => writeln!(w, "GET_LINE {}", prompt)?,
 
+            GetHidden(prompt) => writeln!(w, "GET_HIDDEN {}", prompt)?,
+
             GotIt => writeln!(w, "GOT_IT")?,
 
+            InquireMaxLen(l) => writeln!(w, "INQUIRE_MAXLEN {}", l)?,
+
             Failure {
                 location,
                 error,
             } => {
-                writeln!(w, "FAILURE {} {}", location, *error as isize)?;
+                writeln!(w, "FAILURE {} {}", location, *error as u32)?;
             },
 
             Unexpected(reason) => {
